# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reusable utility functions for executing methods in parallel.

This will create multiple processes to execute device operations on multiple devices. Device
interactions will be logged to the provided logger. Errors can optionally be raised if device
methods fail.

Example usage:
    sample_function_results = parallel_utils.parallel_process(
        "sample_function",
        self._sample_function,
        device_instances)

    for result in sample_function_results:
        do_something(result)
"""
import multiprocessing
import time
from gazoo_device import gdm_logger
from gazoo_device.utility import common_utils
from six.moves import queue

logger_gdm = gdm_logger.get_gdm_logger()

TIMEOUT_PROCESS = 600.0
TIMEOUT_TERMINATE_PROCESS = 5
REQUIRED_PROPS = ["name", "DEVICE_TYPE"]


def parallel_process(action_name,
                     fcn,
                     devices,
                     logger=None,
                     parameter_dicts=None,
                     timeout=TIMEOUT_PROCESS):
    """Concurrently apply function to each device.

    TODO(b/152442892): Increase the flexibility of this function.

    Args:
        action_name (str): terse description of the function
        fcn (function): function to execute in parallel
        devices (list): list of Gazoo device objects
        logger (logger): logger object that will be passed to fcn
        parameter_dicts (dict): of arguments to send to the fcn by device_type
        timeout (int): seconds before terminating all the parallel processes

    Returns:
        list: results from parallel functions if return_results flag is specified. this will be
            a list of either an exception raised while running the function or data added
            manually to the multiprocessing queue.

    Raises:
        RuntimeError: if any of the parallel functions raise error or timeouts and
            return_results flag is not true
        AttributeError: if provided devices are missing required props.
    """
    if not parameter_dicts:
        parameter_dicts = {}

    # verify list of devices was recieved
    if not isinstance(devices, list):
        raise RuntimeError(
            "Devices should be a list. Instead its a {}. Devices: {}".format(
                type(devices), devices))

    # verify devices have required properties
    for prop in REQUIRED_PROPS:
        if not all(hasattr(device, prop) for device in devices):
            raise AttributeError(
                "Devices must have {} property. Devices: {}".format(prop, devices))

    device_names = [device.name for device in devices]
    logger_gdm.info(
        "Executing {} concurrently for {}s on devices {}".format(action_name,
                                                                 timeout,
                                                                 ",".join(device_names)))

    # create queues to manage parallel processes
    return_queue = multiprocessing.Queue()
    error_queue = multiprocessing.Queue()
    processes = []

    # initiate new process for each device with provided arguments
    for device in devices:
        device_type = device.DEVICE_TYPE
        args = [fcn, return_queue, error_queue]
        kwargs = {"device": device}
        if logger:
            kwargs["logger"] = logger
        if device_type in parameter_dicts:
            kwargs["parameter_dict"] = parameter_dicts[device_type]
        process = multiprocessing.Process(target=_sub_process, args=args, kwargs=kwargs)
        processes.append(process)

    # run each process in parallel
    deadline = time.time() + timeout
    for process in processes:
        common_utils.run_before_fork()
        process.start()
        common_utils.run_after_fork_in_parent()

    for process in processes:
        remaining_timeout = max(0, deadline - time.time())  # ensure timeout >= 0
        process.join(timeout=remaining_timeout)
        if process.is_alive():
            process.terminate()
            process.join(timeout=TIMEOUT_TERMINATE_PROCESS)

    # handle results from queue
    errors = get_messages_from_queue(error_queue)
    if errors:
        raise RuntimeError(", ".join(errors))
    return get_messages_from_queue(return_queue)


def issue_devices_parallel(method_name, devices, parameter_dicts=None, timeout=TIMEOUT_PROCESS):
    """Concurrently issue a device command to multiple devices.

    Args:
        method_name (str): device method to call in parallel.
        devices (list): list of Gazoo Device objects.
        parameter_dicts (dict): of arguments to send to the method by device_type.
        timeout (int): seconds before terminating all the parallel processes.

    Returns:
        list: results of parallel method calls.

    Note:
        The nested function _issue_devices_parallel will be the function executed in separate
        processes for each device.
    """
    def _issue_devices_parallel(device, parameter_dict=None):
        """Execute device method."""
        parameter_dict = parameter_dict or {}
        method = getattr(device, method_name)
        return method(**parameter_dict)

    return parallel_process(method_name,
                            _issue_devices_parallel,
                            devices,
                            parameter_dicts=parameter_dicts,
                            timeout=timeout)


def get_messages_from_queue(mp_queue, timeout=0.01):
    """Safely get all messages from a multiprocessing queue.

    Args:
        mp_queue (queue): a multiprocess Queue instance
        timeout (float): seconds to block other processes out from the queue

    Returns:
        list: List of messages or an empty list if there weren't any
    """
    msgs = []
    # According to the python docs https://docs.python.org/2/library/multiprocessing.html
    # after putting an object on an empty queue there may be an
    #  infinitesimal delay before the queue's empty() method returns False
    #
    # We've actually run into this (SPT-1354) so we'll first kick the
    # tires with a get() and then see if there are more using empty()
    try:
        msgs.append(mp_queue.get(True, timeout))
    except queue.Empty:
        pass
    else:
        while not mp_queue.empty():
            msgs.append(mp_queue.get_nowait())
    return msgs


def _sub_process(fcn, return_queue, error_queue, **kwargs):
    """Helper function for calling methods in parallel.

    Args:
        fcn (function): function to execute
        return_queue (queue): a multiprocess Queue instance for results
        error_queue (queue): a multiprocess Queue instance for exceptions
        **kwargs (dict): other keyworded arguments to pass to the function
    """
    try:
        result = fcn(**kwargs)
        if result is not None:
            return_queue.put_nowait(result)
    except Exception as err:
        error_queue.put_nowait(repr(err))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for working with USB mass storage device.

This module uses a combination of ``pyudev`` and ``psutil`` to list and filter
USB mass storage devices.

.. note:

    According to the udiskctl man page
    http://manpages.ubuntu.com/manpages/trusty/man8/udisks.8.html#contenttoc5

    Therefore, applications can not rely on tools like e.g.  udisksctl(1) to be
    available. Additionally, there is no guarantee that the options,
    command-line switches etc. of command-line tools or similar will remain
    stable.

    Instead, applications should only use the D-Bus API, the libudisks2
    library or tools such as dbus-send(1) or gdbus(1) to interact with
    udisksd(8).
"""
from __future__ import absolute_import
import os
import re
import shutil
import stat
import subprocess
import sys
import time
import psutil
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.utility import usb_utils


logger = gdm_logger.get_gdm_logger()


def _validate_suid_bit(cmd_str):
    if sys.platform == "darwin":
        return
    binary = cmd_str.split()[0]
    if not os.stat(binary).st_mode and stat.S_ISUID:
        raise errors.GazooDeviceError("Missing SUID bit on {binary} "
                                      "Fix using 'sudo chmod u+s {binary}'".format(binary=binary))


class UsbStorage(object):
    """USB mass storage device class.

    Used in 'with block' and will mount the USB mass storage device upon entering the with
    block and eject the USB mass storage device upon exiting the 'with block'. If
    the UsbStorage class is not used within a 'with block' then the methods of
    mount, unmount, and eject can be called separately.
    """
    commands_dict = {
        "darwin": {
            "EJECT": 'diskutil eject {}',
            "MOUNT": 'diskutil mount {}',
            "UNMOUNT": 'diskutil unmount {}'
        },
        "linux": {
            "EJECT": '/usr/bin/eject {}',
            "MOUNT": '/usr/bin/udisksctl mount --no-user-interaction -b {}',
            "UNMOUNT": '/bin/umount {}'
        }}

    def __init__(self, serial_number, add_log_note_func=None, eject_on_exit=True):
        """Initialize the USB mass storage device instance with the information specified.

        Args:
            serial_number (str): serial number to identify device.
            add_log_note_func (func): function to log usb serial methods to device logs.
            eject_on_exit (bool): whether to eject or unmount on exit from context manager.
        """
        self._serial_number = serial_number
        self.add_log_note_func = add_log_note_func
        self.eject_on_exit = eject_on_exit
        if sys.platform == "darwin":
            self.commands = self.commands_dict["darwin"]
        else:
            self.commands = self.commands_dict["linux"]

    def __enter__(self):
        self.mount()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.eject_on_exit:
            self.eject()
        else:
            self.unmount()

    @property
    def disk(self):
        """Disk that device mounted to. Empty if not mounted."""
        devices_info = self._get_devices_info()
        return devices_info[0].disk if devices_info else ""

    @property
    def mount_point(self):
        """Returns the mass storage mount point to the caller.

        Returns:
            str: mount path location for caller to use in opening files on the mass storage
                    device
        """
        usb_endpoint_infos = self._get_devices_info()
        if usb_endpoint_infos:
            partitions = psutil.disk_partitions()
            for usb_info_inst in usb_endpoint_infos:
                for partition in partitions:
                    if partition.device == usb_info_inst.disk:
                        return partition.mountpoint
        return None

    def add_log_note(self, msg):
        """Adds note in device logs if method available."""
        if self.add_log_note_func:
            self.add_log_note_func(msg)

    def copy_file_to(self, source_file_path, relative_destination_dir=None):
        """Copies source file to destination directory specified.

        Args:
            source_file_path (str): path to source file to copy to mass storage device
            relative_destination_dir (str): optional path relative to the mount path of the mass
                                            storage device

        Raises:
            GazooDeviceError: if storage is not mounted, the mount path doesn't exist,
                             or if the source file doesn't exist after copying the source
                             file to the mount path + relative destination directory destination

        Note:
            The storage device must be mounted before calling this method. If the relative
            destination directory specified doesn't exist it will be created before the
            source file is copied
        """
        if not self.is_mounted():
            raise errors.GazooDeviceError("Device copy file failed. "
                                          "Device with Serial {} is not mounted.".
                                          format(self._serial_number))

        mount_point = self.mount_point
        if relative_destination_dir is not None:
            dst_file_path = os.path.join(mount_point, relative_destination_dir)
        else:
            dst_file_path = mount_point

        if not os.path.exists(dst_file_path):
            logger.debug("Creating destination path %s", dst_file_path)
            os.makedirs(dst_file_path)

        file_name = os.path.basename(source_file_path)
        dst_file_path = os.path.join(dst_file_path, file_name)
        msg = "Copying {} to {}".format(source_file_path, dst_file_path)
        logger.debug(msg)
        self.add_log_note(msg)
        shutil.copyfile(source_file_path, dst_file_path)
        if not os.path.exists(dst_file_path):
            raise errors.GazooDeviceError("Device copy file failed. "
                                          "File does not exist at destination {}"
                                          "Check destination path permissions.".
                                          format(dst_file_path))
        msg = "Copying {} to {} completed".format(source_file_path, dst_file_path)
        logger.debug(msg)
        self.add_log_note(msg)

    def eject(self, timeout=100.0):
        """Unmounts and ejects the mass storage device within the timeout specified.

        Args:
            timeout (float): time to wait to confirm the mass storage has been unmounted and
                           ejected

        Raises:
            GazooDeviceError: if timeout is exceeded before storage is confirmed as being
                             unmounted or ejected.
        """
        start_time = time.time()
        self.add_log_note("{} eject requested.".format(self._serial_number))
        if self.is_device_ejected(timeout=.1):
            self.add_log_note("{} already ejected.".format(self._serial_number))
            return
        self.unmount(timeout=timeout)
        _validate_suid_bit(self.commands["EJECT"])
        try:
            cmd_list = self.commands["EJECT"].format(self.disk).split()
            subprocess.check_output(cmd_list)
        except subprocess.CalledProcessError as err:
            raise errors.GazooDeviceError("Device eject failed. "
                                          "{} produced this error: {!r}".format(
                                              cmd_list,
                                              err.output))
        if self.is_device_ejected(timeout=timeout):
            self.add_log_note("{} ejected in {}s".format(
                self._serial_number, int(time.time() - start_time)))
            return
        raise errors.GazooDeviceError("Device eject failed. "
                                      "Unable to eject USB mass storage {} in {}s"
                                      "Check if device is currently in use. ".
                                      format(self.mount_point, timeout))

    def is_device_ejected(self, timeout=20):
        """Check whether device is ejected from the system.

        Args:
            timeout (int):  Number of seconds to wait for successful eject.
        Returns:
            bool: True if device is ejected.
        """
        exit_time = time.time() + timeout
        while time.time() < exit_time:
            devices_info = self._get_devices_info()
            if not devices_info:
                return True
            if sys.platform != "darwin":
                device_id_path = devices_info[0].disk
                with open('/etc/fstab', 'r') as open_file:
                    text = open_file.read()
                if not re.search(device_id_path, text):
                    return True
            time.sleep(.5)
        return False

    def is_found(self):
        """Returns True if the mass storage device has been detected and is available to mount.

        Returns:
            bool: True if mass storage device is detected
        """
        return bool(self._get_devices_info())

    def is_mounted(self, timeout=20):
        """Returns True if the mass storage device is currently mounted.

        Args:
            timeout (float): time to wait for it to mount before timeout.
        Returns:
            bool: True if mass storage device is mounted

        Note:
           Uses a timeout as not all devices are using mount to mount the device and it
           can take a few seconds for the computers to register the external mount.
        """
        start_time = time.time()
        exit_time = start_time + timeout
        while time.time() < exit_time:
            if self.mount_point:
                msg = "{} found mounted at {} in {}s".format(
                    self._serial_number, self.mount_point, int(time.time() - start_time))
                self.add_log_note(msg)
                logger.debug(msg)
                return True
        return False

    def mount(self, timeout=180):
        """Mounts the mass storage device within the timeout specified.

        Args:
            timeout (float): time to wait to confirm storage has been mounted and/or after
                             manually attempting to mount the storage

        Raises:
            GazooDeviceError: if timeout is exceeded before storage is confirmed as
            being mounted.

        Note:
            Since there is a udev rule that is responsible for performing the actual mounting of
            the device this method will wait up to the timeout specified before attempting to
            manually mount the device and then wait again for the timeout specified to confirm
            that the device has been successfully mounted
        """
        self.add_log_note("{} mount requested.".format(self._serial_number))
        if self.is_mounted():
            return
        if not self.is_found():
            self.add_log_note("Error: {} not found on machine.".format(self._serial_number))
            raise errors.GazooDeviceError("Device mount failed. "
                                          "USB mass storage matching Serial#={} was not found. "
                                          "\nCheck cables and device or try rebooting.".
                                          format(self._serial_number))
        msg = "Mounting {}".format(self.disk)
        logger.debug(msg)
        self.add_log_note(msg)
        _validate_suid_bit(self.commands["MOUNT"])
        try:
            cmd_list = self.commands["MOUNT"].format(self.disk).split()
            subprocess.check_output(cmd_list)
        except subprocess.CalledProcessError as err:
            if isinstance(err.output, bytes):
                err.output = err.output.decode('utf-8')
            if 'Not authorized' in err.output:
                err = "Try running 'sudo chmod u+s {}' to fix permission error.".format(
                    ' '.join(cmd_list[0]))
            raise errors.GazooDeviceError("Device mount failed. Err: {!r}".format(err))

        if not self.is_mounted(timeout=timeout):
            raise errors.GazooDeviceError("Device mount failed. Not mounted after {}s".format(
                timeout))

    def move_file_from(self, relative_source_file_path, destination, delete_file=True):
        """Moves file from storage mount point to destination directory specified.

        Args:
            relative_source_file_path (str): relative path to file on device to copy over.
            destination (str): destination directory or file name path
            delete_file (bool): delete file from device.

        Raises:
            GazooDeviceError: if storage is not mounted, the mount path doesn't exist,
                             or if the source file doesn't exist

        Note:
            The storage device must be mounted before calling this method.

        """
        if not self.is_mounted():
            raise errors.GazooDeviceError("Move_file_from failed. "
                                          "Device with Serial {} is not mounted.".
                                          format(self._serial_number))

        source_file_path = os.path.join(self.mount_point,
                                        relative_source_file_path)

        if not os.path.exists(source_file_path):
            raise errors.GazooDeviceError("Device copy file failed. "
                                          "Source file {} doesn't exist. ".
                                          format(source_file_path))

        msg = "Copying {} to {}".format(source_file_path, destination)
        logger.debug(msg)
        self.add_log_note(msg)
        shutil.copy(source_file_path, destination)

        if delete_file:
            logger.debug("Removing %s", source_file_path)
            os.remove(source_file_path)

        if not os.path.exists(destination):
            raise errors.GazooDeviceError("Device copy file failed. "
                                          "Destination {} does not exist. "
                                          "Check destination path permissions.".
                                          format(destination))
        msg = "Copying {} to {} completed.".format(relative_source_file_path, destination)
        logger.debug(msg)
        self.add_log_note(msg)

    def unmount(self, timeout=180.0):
        """Unmounts the mass storage device within the timeout specified.

        Args:
            timeout (float): time to wait to confirm storage and successfully unmounted

        Raises:
            GazooDeviceError: if timeout is exceeded before storage is confirmed as being
                             unmounted
        """
        start_time = time.time()
        exit_time = start_time + timeout
        self.add_log_note("{} unmount requested".format(self._serial_number))
        mount_point = self.mount_point
        if not mount_point:
            self.add_log_note("{} already unmounted".format(self._serial_number))
            return
        self.add_log_note(
            "{} unmounting from {}".format(
                self._serial_number,
                mount_point))

        _validate_suid_bit(self.commands["UNMOUNT"])
        try:
            cmd_list = self.commands["UNMOUNT"].format(mount_point).split()
            subprocess.check_output(cmd_list)
        except subprocess.CalledProcessError as err:
            raise errors.GazooDeviceError("Device unmount failed. "
                                          "Error when calling /bin/umount: {}".format(
                                              err.output))

        while time.time() < exit_time:
            if not self.mount_point:
                msg = "{} successfully unmounted in {}s".format(
                    self._serial_number,
                    int(time.time() - start_time))
                logger.debug(msg)
                self.add_log_note(msg)
                return

        raise errors.GazooDeviceError("Device unmount failed. "
                                      "Unmounting USB mass storage at {} failed after {}s".
                                      format(self.mount_point, timeout))

    def _get_devices_info(self):
        """Returns info for the device.

        Returns:
          list: one or more matching usbinfo objects for each usb endpoint that matches.
        """
        return usb_utils.find_matching_connections(
            {"serial_number": {"include_regex": self._serial_number}})



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the switch_power_usb_with_charge capability."""
from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities import switch_power_usb_default

logger = gdm_logger.get_gdm_logger()

CHARGE = "charge"
OFF = "off"
SYNC = "sync"


class SwitchPowerUsbWithCharge(switch_power_usb_default.SwitchPowerUsbDefault):
    """Definition of the switch_power_usb_with_charge capability."""
    _REQUIRED_COMMANDS = ["SET_MODE", "GET_MODE"]
    _REQUIRED_REGEXS = ["GET_MODE_REGEX"]

    def __init__(self,
                 shell_fn,
                 regex_shell_fn,
                 command_dict,
                 regex_dict,
                 device_name,
                 serial_number,
                 total_ports):
        """Create an instance of the hub power with charge capability.

        Args:
            shell_fn (func): Function for calling shell on the device.
            regex_shell_fn (func): Function for calling shell_with_regex on the device.
            command_dict (dict): A dictionary containing the command used for each property.
                                 The dictionary must contain the following keys:

                                 - SET_MODE
                                 - GET_MODE

            regex_dict (dict): A dictionary containing the response regex to use to
                               determine the return value for the property.
                               The dictionary must contain the following keys:

                               - GET_MODE_REGEX

            device_name (str): name of the device this capability is attached to.
            serial_number (str): serial number of device this capability is attached to.
            e.g. {'OFF': 'off', 'POWER_ON': 'sync', 'POWER_CHARGE': 'charge'}
            total_ports (int): number of ports on device
        """
        super().__init__(shell_fn=shell_fn,
                         regex_shell_fn=regex_shell_fn,
                         command_dict=command_dict,
                         regex_dict=regex_dict,
                         device_name=device_name,
                         serial_number=serial_number,
                         total_ports=total_ports)

    @decorators.PersistentProperty
    def supported_modes(self):
        """Get the USB power modes supported by the USB hub."""
        return [OFF, SYNC, CHARGE]

    def get_mode(self, port):
        """Gets the USB mode for the specified port.

        Args:
          port (int): Use this port to get the mode.

        Returns:
            str: USB port mode settings
                 Example, 'off', 'sync', 'charge'

        Raises:
          GazooDeviceError: invalid port.
        """
        port = int(port)
        self._validate_port("get_mode", port)
        flags = self._regex_shell_fn(self._command_dict["GET_MODE"].format(port),
                                     self._regex_dict["GET_MODE_REGEX"],
                                     tries=5)
        if "O" in flags:
            mode = OFF
        elif "S" in flags:
            mode = SYNC
        else:
            mode = CHARGE
        return mode

    @decorators.CapabilityLogDecorator(logger, decorators.DEBUG)
    def power_off(self, port):
        """This command powers off the port specified.

        Args:
            port (int): Identifies which hub port to power off.
        """
        port = int(port)
        self._validate_port("power_off", port)
        self.set_mode(OFF, port)

    @decorators.CapabilityLogDecorator(logger, decorators.DEBUG)
    def power_on(self, port, data_sync=True):
        """This command powers on the port specified.

        Args:
            port (int): Identifies which hub port to power on.
            data_sync (bool): True if data should be enabled, false for power only
        """
        port = int(port)
        self._validate_port("power_on", port)
        if data_sync:
            self.set_mode(SYNC, port)
        else:
            self.set_mode(CHARGE, port)

    @decorators.CapabilityLogDecorator(logger, decorators.DEBUG)
    def set_mode(self, mode, port):
        """Sets the given USB port to the mode specified.

        Args:
          mode (str): USB mode to set. Options: 'off', 'sync', 'charge'
          port (int): The port to set.

        Raises:
          GazooDeviceError: invalid port, or mode.
        """
        port = int(port)
        self._validate_port("set_mode", port)
        self._validate_mode(mode)
        logger.debug("{} setting power mode to {} for usb port {}".
                     format(self._device_name, mode, port))
        self._shell_fn(self._command_dict["SET_MODE"].format(mode, port))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for analyzing GDM code at runtime (reflection)."""
import functools
import importlib
import inspect
import os

import gazoo_device

from gazoo_device import auxiliary_devices
from gazoo_device import config
from gazoo_device import gdm_logger
from gazoo_device import primary_devices
from gazoo_device import virtual_devices
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.base_classes import gazoo_device_base
from gazoo_device.capabilities.interfaces import capability_base
from gazoo_device.utility import common_utils
from gazoo_device.utility import supported_classes


logger = gdm_logger.get_gdm_logger()

_PACKAGE_DIR = "gazoo_device/"


def identify_all_supported_classes():
    """Dynamically identify all device classes and capabilities supported by GDM.

    The results are stored in supported_classes.py to avoid circular imports.
    """
    if supported_classes.mappings_generated:
        return

    aux_classes = _get_all_device_classes(auxiliary_device.AuxiliaryDevice,
                                          auxiliary_devices)

    primary_classes = _get_all_device_classes(gazoo_device_base.GazooDeviceBase,
                                              primary_devices)

    virtual_classes = _get_all_device_classes(gazoo_device_base.GazooDeviceBase,
                                              virtual_devices)

    supported_classes.SUPPORTED_AUXILIARY_DEVICE_CLASSES = aux_classes
    supported_classes.SUPPORTED_PRIMARY_DEVICE_CLASSES = primary_classes
    supported_classes.SUPPORTED_VIRTUAL_DEVICE_CLASSES = virtual_classes

    supported_classes.SUPPORTED_CAPABILITY_INTERFACES = {
        common_utils.generate_name(interface): interface
        for interface in _get_all_supported_capability_interfaces()}

    supported_classes.SUPPORTED_CAPABILITY_FLAVORS = {
        common_utils.generate_name(flavor): flavor
        for flavor in _get_all_supported_capability_flavors()}

    supported_classes.SUPPORTED_CAPABILITIES = {
        interface.get_capability_name(): if_name
        for if_name, interface in supported_classes.SUPPORTED_CAPABILITY_INTERFACES.items()}

    supported_classes._mappings_generated = True


def get_all_subclasses_in_package(parent_class, package,
                                  excluded_modules=None, only_public=True):
    """Returns all classes derived from parent_class in any of the modules within the package.

    Does not recurse into subpackages (subdirectories will not be scanned).

    Args:
        parent_class (type): class object from which the subclasses must be derived.
        package (module): module (__init__.py) of package to scan.
        excluded_modules (list): list of modules to exclude from the scan.
        only_public (bool): whether to return only public or both public and private classes.

    Returns:
        dict: map from module to set of classes derived from parent_class defined in that module.
    """
    if excluded_modules is None:
        excluded_modules = set()

    modules = [module for module in _list_package_modules(package)
               if module not in excluded_modules]

    module_to_classes = {}
    for module in modules:
        public_classes = [member for name, member in inspect.getmembers(module)
                          if inspect.isclass(member) and
                          (not name.startswith("_") or not only_public)]
        derived_classes = [a_class for a_class in public_classes
                           if issubclass(a_class, parent_class)]
        if derived_classes:
            module_to_classes[module] = set(derived_classes)
    return module_to_classes


def _get_all_device_classes(base_class, package):
    """Dynamically identify all device classes using the base class in the package.

    Args:
       base_class (class): base class of device class
       package (module): package in which classes reside

    Note:
        - device classes must be in package folder;
        - all device classes must inherit from base class;
        - subdirectories of package folder will not be scanned.

    Returns:
        list: all device classes using the base class in the package.
    """
    device_class_list = []
    final_device_classes = []
    classes_superset = get_all_subclasses_in_package(base_class, package).values()
    for device_class_set in classes_superset:
        device_class_list += list(device_class_set)

    for device_class in device_class_list:
        class_name = _lowercase_object_name(device_class)
        if class_name.endswith("device") or class_name.endswith("base"):
            # remove any imported base classes
            continue
        else:
            final_device_classes.append(device_class)
    final_device_classes.sort(key=lambda x: x.DEVICE_TYPE)
    return final_device_classes


def _get_all_supported_capability_flavors():
    """Dynamically identify all capability flavors supported by GDM.

    Note:
        - capability flavors must be in gazoo_device/capabilities/ folder;
        - subdirectories of gazoo_device/capabilities/ will not be scanned.
          In other words, there must be no subfolders under gazoo_device/capabilities/.
        - capability flavors must (indirectly) inherit from CapabilityBase. Capability flavors
          should inherit from capability interfaces, which should inherit from CapabilityBase.

    Returns:
        set: all capability flavor classes supported by GDM.
    """
    flavor_module_to_classes = get_all_subclasses_in_package(
        capability_base.CapabilityBase, gazoo_device.capabilities)
    return functools.reduce(set.union, flavor_module_to_classes.values(), set())


def _get_all_supported_capability_interfaces():
    """Dynamically identify all capability interfaces supported by GDM.

    Note:
        - capability interfaces (abstract base classes) must be in
          gazoo_device/capabilities/interfaces/ folder;
        - subdirectories of gazoo_device/capabilities/interfaces/ will not be scanned.
          In other words, there must be no subfolders under gazoo_device/capabilities/interfaces/.
        - capability interfaces must inherit from CapabilityBase;
        - CapabilityBase class is excluded from the results.

    Returns:
        set: all capability interface classes supported by GDM.
    """
    interface_module_to_classes = get_all_subclasses_in_package(
        capability_base.CapabilityBase, gazoo_device.capabilities.interfaces,
        excluded_modules=[capability_base])
    return functools.reduce(set.union, interface_module_to_classes.values(), set())


def _get_module_path(module):
    """Return the module path.

    Args:
        module(module): python module object.

    Returns:
        str: path to the module.
    """
    module_path = module.__file__
    if module_path.startswith(config.PACKAGE_PATH):
        # No action necessary if root directories are the same
        return module_path
    else:
        if _PACKAGE_DIR in module_path:  # "<...path...>/gazoo_device/<...path remainder...>"
            path_remainder = module_path.split(_PACKAGE_DIR)[-1]
            return os.path.join(config.PACKAGE_PATH, path_remainder)
        else:  # Not a gazoo_device module.
            return module_path


def _list_package_modules(package):
    """List all modules defined in a package. Subpackages (subdirectories) are not scanned.

    Args:
        package(module): package to scan. For example, gazoo_device.capabilities.

    Returns:
        list: all modules that can be imported from the package.
    """
    suffix = ".py"
    suffix_len = len(suffix)

    package_path = _get_module_path(package)
    package_dir = os.path.dirname(package_path)
    files_in_package = os.listdir(package_dir)

    module_names = [a_file[:-suffix_len] for a_file in files_in_package
                    if a_file.endswith(suffix) and a_file != "__init__.py"]

    modules = []
    for module_name in module_names:
        try:
            modules.append(importlib.import_module(".{}".format(module_name),
                                                   package=package.__name__))
        except ImportError:
            logger.debug("Failed to import module {} from {}.".format(module_name,
                                                                      package.__name__))
    return modules


def _lowercase_object_name(a_object):
    return a_object.__name__.lower()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DataFramer based classes transforms incoming raw data into partial and complete lines.

The NewlineFramer (default) splits incoming raw data into partial and complete lines
at newline character boundaries.

The InterwovenLogFramer finds and extracts log lines interwoven with responses
using the log line regex provided.

Notes:
A complete line returned by all DataFramer based classes
should include the trailing newline
character as is.

Partial lines are typically those raw bytes that remain which
don't have a trailing newline character.
"""
from __future__ import absolute_import
import re
import six
from six.moves import range
import abc


class DataFramer(six.with_metaclass(abc.ABCMeta), object):
    """Abstract class for transforming raw device input into discreet lines."""

    @abc.abstractmethod
    def get_lines(self, raw_data, begin=0, end=None):
        """Yields each line from the raw_data provided from begin to end.

        Args:
            raw_data (str): to split into framed lines
            begin (int): index to the first character to start looking for newlines
            end (int): index to the last character to look for newline characters

        Yields:
            str: A partial or complete framed line found in the raw_data provided.

        Note:
            The begin argument allows for a slight optimization by skipping
            past previous partial line in search of the next newline character.
        """


def _validate_end_and_begin_are_valid(begin, end, string_len):
    """Validate begin and end are in the right range and are integers.

    Args:
        begin (int): character position to start searching for newlines from.
        end (int): end character position to stop searching for newlines from.
        string_len (int): length of the string to search for newlines.

    Raises:
       ValueError: if begin or end are not valid intput.
    """
    if not isinstance(begin, int):
        raise ValueError("Expected begin to be an integer found {}".
                         format(type(begin)))
    elif begin < 0 or begin > string_len:
        raise ValueError("Expected begin value to be >= 0 and <= {}, found {}".
                         format(string_len, begin))
    elif end is not None:
        if not isinstance(end, int):
            raise ValueError("Expected end to be an integer found {}".
                             format(type(end)))
        elif end < 0 or end > string_len:
            raise ValueError("Expected end value to be >= 0 and <= {}, found {}".
                             format(string_len, end))
        elif begin > end:
            raise ValueError("Expected end value {} to be <= begin value {}".
                             format(end, begin))


def split_newlines_only(string, begin=0, end=None, keepends=True, cleanends=False):
    """Generates substrings of string up to and including newline characters only.

    Args:
        string (str): to extract substrings from.
        begin (int): character to start searching for newlines from, not the
                     first character of the first substring returned.
        end (int): end character to stop searching for newlines from
        keepends (bool): flag indicating newline character should be returned
                         as part of substring returned.
        cleanends (bool): flag indicating trailing line return and line feeds
                          should be stripped off and replaced with a line feed.

    Raises:
        ValueError: if begin or end < 0 or > than length of the string

    Yields:
        str: Generates substrings of string from 0 to len(string) for each
             newline character found and optionally including the newline
             character.

    Note:
        The begin argument allows for skipping characters that you know won't
        contain newline characters. The first substring returned will always be
        from the first character of the string to the first newline character
        found or the end of the string if no newline characters exist.
        If the end argument is None then the length of the string will be used
        instead. If the end argument is specified and is less than the length
        of the string then this function will yield all characters from end to
        the length of the string as the last substring.
    """
    string_len = len(string)
    _validate_end_and_begin_are_valid(begin, end, string_len)
    if end is None:
        end = string_len
    start_index = 0
    if cleanends:
        pre_line_return = begin
        # If begin > 0 might need to find pre_line_return spot in string
        for i in range(begin, 0, -1):
            if string[i] != "\r" and string[i] != "\n":
                pre_line_return = i + 1
                break
    end_offset = 1 if keepends else 0
    for i in range(begin, end):
        if cleanends and string[i] != "\r" and string[i] != "\n":
            pre_line_return = i + 1
        if string[i] == "\n":
            if cleanends:
                yield string[start_index:pre_line_return] + string[i:i + end_offset]
            else:
                yield string[start_index:i + end_offset]
            start_index = i + 1
            if start_index == end:
                break
    else:
        yield string[start_index:]


class InterwovenLogFramer(DataFramer):
    """Finds and extract out of order logs that are interwoven with responses."""

    def __init__(self, log_line_pattern, add_newline=False, keepends=True, cleanends=True):
        """Initializes LogFramer with the log line pattern provided.

        Args:
            log_line_pattern (str): to use to match and extract log lines from
                                    response lines.
            add_newline (bool): add a newline character to the extracted log line
            keepends (bool): flag indicating that trailing LF should be returned
            cleanends (bool): flag indicating that trailing CR and LF should be
                              replaced with LF only.

        Note:
            The log_line_pattern should contain a single regular expression
            group that contains the complete log line, including the newline
            character (or use add_newline to add one), to be extracted from the
            response line.

            The cleanends flag will strip off all CR and LF characters at the
            end of the line. If keepends is True the LF character will be
            returned. This is done before the log_line_pattern logic is
            performed so that add_newline only adds a LF character to the
            log_line string matched if enabled.
        """
        self.add_newline = add_newline
        self.keepends = keepends
        self.cleanends = cleanends
        self.log_line_re = re.compile(log_line_pattern)

    def get_lines(self, raw_data, begin=0, end=None):
        """Yields each deframed line from the raw_data provided from begin to end.

        Args:
            raw_data (str): to split into framed lines
            begin (int): index to the first character to start looking for newlines
            end (int): index to the last character to look for newline characters

        Yields:
            str: A partial or complete framed line found in the raw_data provided.

        Note:
            The begin argument allows for a slight optimization by skipping
            past previous partial line in search of the next newline character.
        """

        for raw_line in split_newlines_only(raw_data,
                                            begin=begin,
                                            end=end,
                                            keepends=self.keepends,
                                            cleanends=self.cleanends):
            if raw_line[-1] != "\n":
                yield raw_line
            else:
                match = self.log_line_re.search(raw_line)
                if match:
                    line1 = match.group(1)
                    if self.add_newline:
                        line1 += u"\n"
                    yield line1
                    line2 = raw_line[:match.start()] + raw_line[match.end():]
                    if line2:
                        yield line2
                else:
                    yield raw_line


class NewlineFramer(DataFramer):
    """Splits lines at each newline character.

    Note:
        This is the default DataFramer.
    """

    def __init__(self, keepends=True, cleanends=True):
        """Initializes NewlineFramer with the options specified.

        Args:
            keepends (bool): flag indicating that trailing LF should be returned
            cleanends (bool): flag indicating that trailing CR and LF should be
                              replaced with LF only.

        Note:
            The cleanends flag will strip off all CR and LF characters at the
            end of the line. If keepends is True the LF character will be
            returned. No LF characters are added in the process.
        """
        self.keepends = keepends
        self.cleanends = cleanends

    def get_lines(self, raw_data, begin=0, end=None):
        """Returns each deframed line from the raw_data provided from begin to end.

        Args:
            raw_data (str): to split into framed lines
            begin (int): index to the first character to start looking for newlines
            end (int): index to the last character to look for newline characters

        Returns:
            generator: A generator function that can be used to split raw_data
                       into partial or complete framed lines on newline
                       characters.

        Note:
            The begin argument allows for a slight optimization by skipping
            past previous partial line in search of the next newline character.
        """

        return split_newlines_only(raw_data,
                                   begin=begin,
                                   end=end,
                                   keepends=self.keepends,
                                   cleanends=self.cleanends)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface containing required GDM APIs for first party devices.

Some of these APIs are implemented in GazooDeviceBase and don't need to be re-implemented
in derived device classes. Device classes should inherit from GazooDeviceBase, not from this class.
"""
import abc


class FirstPartyDeviceBase(abc.ABC):
    """Interface containing required GDM APIs for first party devices."""
    COMMUNICATION_TYPE = None  # Override
    DETECT_MATCH_CRITERIA = None  # Overrride
    DEVICE_TYPE = None  # Override
    _COMMUNICATION_KWARGS = {}

    @abc.abstractproperty
    def alias(self):
        """Returns the user-defined device alias (string)."""

    @abc.abstractproperty
    def commands(self):
        """Dictionary of commands issued to the device via shell."""

    @abc.abstractproperty
    def communication_address(self):
        """Returns the name of the main communication port (for example, ip address)."""

    @abc.abstractproperty
    def connected(self):
        """Returns whether the device is connected or not."""

    @abc.abstractproperty
    def firmware_version(self):
        """Returns the firmware version installed on the device."""

    @abc.abstractproperty
    def health_checks(self):
        """Returns list of methods to execute as health checks."""

    @abc.abstractproperty
    def model(self):
        """Returns the device model."""

    @abc.abstractproperty
    def name(self):
        """Returns the unique identifier for the device (such as Cambrionix-z3pv)."""

    @abc.abstractproperty
    def os(self):
        """Os of the device. IE Android, Linux."""

    @abc.abstractproperty
    def platform(self):
        """Platform of the device."""

    @abc.abstractproperty
    def regexes(self):
        """Regular expressions used to retrieve properties, events, states from device output."""

    @abc.abstractproperty
    def serial_number(self):
        """Returns the serial number of the device."""

    @abc.abstractproperty
    def timeouts(self):
        """Dictionary of default timeouts to use when expecting certain actions."""

    @abc.abstractmethod
    def check_device_ready(self):
        """Checks if the device is ready for testing. If not, raises a CheckDeviceReadyError.

        This is typically implemented by checking if device responds to shell commands and if
            it's streaming logs. There can be additional device-specific checks as well.
        """

    @abc.abstractmethod
    def close(self):
        """Calls close on the communication ports and resets anything needed.

        Note:
            Resets the buttons and terminates the child processes.
        """

    @abc.abstractmethod
    def get_detection_info(self):
        """Gets the persistent and optional attributes of a device during setup.

        Returns:
          tuple: (dict, dict) dictionary of persistent attributes,
                  dictionary of optional attributes (set to None).
        """

    @abc.abstractmethod
    def get_dynamic_properties(self):
        """Returns dictionary of dynamic properties."""

    @abc.abstractmethod
    def get_settable_properties(self):
        """Returns dictionary of settable properties."""

    @abc.abstractmethod
    def get_persistent_properties(self):
        """Returns dictionary of persistent properties."""

    @abc.abstractmethod
    def get_property_names(self):
        """Returns a list of all property names."""

    @abc.abstractmethod
    def factory_reset(self):
        """Places the device back in an out-of-box state.

        Note:
            It unpairs the device as necessary.
        """

    @abc.abstractmethod
    def is_connected(cls, device_config):  # pylint: disable=no-self-argument
        """Determines if the device is connected (reachable).

        Note:
            This method is used during device detection.
            This should be a class or static method, not an instance method.

        Args:
            device_config (dict): contains "persistent" dict

        Returns:
            bool: True if the device is connected (i.e. pingable), False otherwise.
        """

    @abc.abstractmethod
    def make_device_ready(self, setting):
        """Validates and puts device in a testable state or raises an error.

        Args:
           setting (str): off | on | check_only

        Notes:
           If setting is off, will skip check_device_ready and recover.
           If check_only, will skip recovery.
        """

    @abc.abstractmethod
    def reboot(self, no_wait=False, method="shell"):
        """Issues a soft reboot command.

        Args:
            no_wait (bool): flag indicating whether reboot verification should be skipped.
                            If False, blocks until reboot completion.
            method (str): reboot technique to use.
        """

    @abc.abstractmethod
    def recover(self, error):
        """Attempts to recover device based on the type of error specified.

        Note: The check_device_ready method can raise a number of separate exceptions which
        are passed to this method as exception objects. The recovery method is chosen based on
        the type of the error. See subclasses of CheckDeviceError in gazoo_device.errors for a
        a list of errors raised by check_device_ready.

        Args:
            error (CheckDeviceReadyError): A subclass of CheckDeviceReadyError that will be
                                           used to identify a possible recovery solution to use.

        Raises:
            GazooDeviceError: If device recovery fails while attempting to perform recovery
                             steps.
            CheckDeviceReadyError: If there are no recovery steps available for the error
                                   argument, it will be re-raised directly.
        """

    @abc.abstractmethod
    def shell(self,
              command,
              command_name="shell",
              timeout=None,
              port=0,
              searchwindowsize=2000,
              include_return_code=False):
        """Sends command and returns response.

        Args:
            command (str): Command to send to the device.
            command_name (str): Optional identifier to use in logs for this command.
            timeout (float): Seconds to wait for pattern after command sent.
            port (int): Which port to send on, 0 or 1. Default: 0.
            searchwindowsize (int): Number of the last bytes to look at
            include_return_code (bool): flag indicating return code should be returned

        Returns:
            str or tuple: response or (response, return_code) if include_return_code is True.
        """

    @abc.abstractmethod
    def upgrade(self,
                build_number=None,
                build_url=None,
                build_file=None,
                forced_upgrade=False,
                latest_verified=False):
        """Upgrades the device to a given build and verifies success.

        Note:
          defaults to last build validated with gazoo_device regression tests.
          See device.build_properties['build_number']

        Args:
            build_number (int): build number with default build info.
            build_url (str): url or gsutil to the .zip file.
            build_file (str): local path to the file.
            forced_upgrade (bool): Upgrade the firmware without checks.
            latest_verified (bool): upgrade to latest verified build.
        """

    @abc.abstractmethod
    def wait_for_bootup_complete(self, timeout=None):
        """Wait until the device finishes booting up and is ready for testing.

        Args:
            timeout (int): max time to wait for the device to finish booting up.

        Raises:
            DeviceNotBootupCompleteError: device failed to finish booting up before the timeout.
        """



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The minimum base class for switchboard subprocesses."""
from __future__ import absolute_import
import contextlib
import multiprocessing
import os
import six.moves.queue
import signal
import socket
import traceback

import psutil

from gazoo_device import gdm_logger
from gazoo_device.utility import common_utils

logger = gdm_logger.get_gdm_logger()


def get_message(queue, timeout=None):
    """Returns next message from queue.

    Args:
        queue (Queue): to retrieve next message from
        timeout (float): time to wait in seconds for incoming message to arrive.

    Raises:
        ValueError: if queue provided is not a valid queue object

    Returns:
        object: The message data or None if no messages were available or
                retrieved within timeout specified.
    """

    if queue is None or not hasattr(queue, "get") or not hasattr(queue, "task_done"):
        raise ValueError("Invalid queue of {} provided".
                         format(type(queue)))

    message = None
    try:
        message = queue.get(block=True, timeout=timeout)
        queue.task_done()
        # manager shutdown or queue empty
    except (IOError, six.moves.queue.Empty, ValueError, socket.error):
        pass
    return message


def put_message(queue, message, timeout=None):
    """Puts message into queue using the timeout specified.

    Args:
        queue (Queue): to add message to
        message (object): to write into the queue
        timeout (float): time to wait in seconds for outgoing message to
                         be accepted into queue.

    Raises:
        ValueError: if queue provided is not valid or has no put method
        Queue.Full: if outgoing queue is full and timeout was provided and was
                    reached before message was sent.
    """
    if queue is None or not hasattr(queue, "put"):
        raise ValueError("Invalid queue of {} provided".
                         format(type(queue)))

    try:
        queue.put(message, block=True, timeout=timeout)
    except (IOError, socket.error):  # manager shutdown
        pass


@contextlib.contextmanager
def _child_process_wrapper(parent_proc, process_name, device_name, exception_queue):
    """Wrapper to ignore interrupts in child processes; the main process handles cleanup."""
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    try:
        yield
    except EOFError:  # parent process has terminated
        pass
    except Exception:
        stack_trace = traceback.format_exc()
        try:
            exception_queue.put_nowait("Device {} raised exception in {}. "
                                       "{}".
                                       format(device_name,
                                              process_name,
                                              stack_trace))
            os.kill(parent_proc.pid, signal.SIGUSR1)
        except (IOError, OSError):  # queue or parent process doesn't exist anymore
            pass


def _parent_is_alive(parent_proc):
    status = False
    try:
        status = parent_proc.status() != psutil.STATUS_ZOMBIE
    except psutil.NoSuchProcess:  # parent process doesn't exist anymore
        pass
    return status


def _process_loop(cls, parent_proc):
    """Child process loop which handles start/stop events and exceptions."""
    with _child_process_wrapper(parent_proc,
                                cls.process_name,
                                cls.device_name,
                                cls._exception_queue):
        try:
            cls._start_event.set()
        except Exception as err:
            stack_trace = traceback.format_exc()
            logger.info("Device {} Process {} error {!r} "
                        "start event error. {}".
                        format(cls.device_name,
                               cls.process_name,
                               err,
                               stack_trace))
        running = cls._pre_run_hook()
        while running and _parent_is_alive(parent_proc):
            try:
                if cls._terminate_event.is_set():
                    cls._terminate_event.clear()
                    break
            except IOError:  # manager shutdown
                break
            running = cls._do_work()
    try:
        cls._stop_event.set()
    except IOError:  # manager shutdown
        pass
    cls._post_run_hook()


class SwitchboardProcess(object):
    """Simplifies creating Switchboard processes.

    Provides the following capabilities:
        * starting and verifying process start
        * stopping and verifying process stop
        * checking if process was started
        * checking if process is still running
        * providing pre and post process loop hooks
        * signaling main process when exceptions are raised in subprocess
        * receiving queue commands with optional data to be processed
        * Simplifying process loop definition for subclasses
    """

    def __init__(self, device_name, process_name, mp_manager, exception_queue,
                 command_queue, valid_commands=None):
        """Initialize SwitchboardProcess with the arguments provided.

        Args:
            device_name (str): of the device for exception error messages
            process_name (str): to use for process name and exception error messages
            mp_manager (multiprocessing.Manager): object to use for creating Events
            exception_queue (Queue): to use for reporting exception traceback
                                     message from subprocess
            command_queue (Queue): to receive commands into
            valid_commands (list): optional list of valid command strings
        """

        self.device_name = device_name
        self.process_name = process_name
        self._command_queue = command_queue
        self._exception_queue = exception_queue
        self._start_event = mp_manager.Event()
        self._stop_event = mp_manager.Event()
        self._terminate_event = mp_manager.Event()
        self._valid_commands = valid_commands or []

    def __del__(self):
        if self.is_started():
            self.stop()

    def start(self):
        """Starts process if process is not already running.

        Raises:
            RuntimeError: if called when process is already running

        Note:
            When calling this method you should use is_started() to check if
            process was previously started to prevent raising an error.
        """

        if not self.is_started():
            parent_proc = psutil.Process(os.getpid())
            self._start_event.clear()
            self._stop_event.clear()
            process = multiprocessing.Process(name=self.process_name,
                                              target=_process_loop,
                                              args=(self, parent_proc))
            common_utils.run_before_fork()
            process.start()
            common_utils.run_after_fork_in_parent()
            start_event_value = self._start_event.wait(timeout=5)
            if not start_event_value:
                raise RuntimeError("Device {} failed to start child process {}. "
                                   "Start event was not set."
                                   .format(self.device_name, self.process_name))
            self._process = process
        else:
            raise RuntimeError("Device {} failed to start child process {}. "
                               "Child process is already running.".
                               format(self.device_name, self.process_name))

    def is_started(self):
        """Returns True if process was started, False otherwise.

        Return:
            bool: True if process was started
        """

        return hasattr(self, "_process")

    def is_running(self):
        """Returns True if process is currently running, False otherwise.

        Return:
            bool: True if process was started and is currently running

        Note:
            Use this method to determine if a process is still running, since
            the process might choose to self exit.
        """

        if self.is_started():
            return self._process.is_alive()
        return False

    def send_command(self, command, data=None):
        """Sends command with optional data provided.

        Args:
            command (str): string to send
            data (object): optional data to include with command.

        Raises:
            ValueError: if command provided is not in the list of valid commands
        """

        if command not in self._valid_commands:
            raise ValueError("Device {} send command to child process {} failed. "
                             "{} is not a valid command in {!r}".
                             format(self.device_name,
                                    self.process_name,
                                    command,
                                    self._valid_commands))
        put_message(self._command_queue, (command, data))

    def stop(self):
        """Stops process if process is running.

        Raises:
            RuntimeError: if called when process is not currently running
            OSError: if a system related exception occurs.

        Note:
            When calling this method you should use is_started() to check if
            process was previously started to prevent raising an error.
        """

        if self.is_started():
            if self._process.is_alive():
                try:
                    self._terminate_event.set()
                except IOError:  # manager shutdown
                    pass
            try:
                stop_event_value = self._stop_event.wait(timeout=5)
                if not stop_event_value:
                    msg = ("Device {} failed to stop child process {}. "
                           "Stop event was not set.").format(self.device_name, self.process_name)
                    logger.error(msg)
                    raise IOError(msg)
            except (IOError, ValueError):  # manager shutdown failed
                pass
            self._process.join(timeout=1)
            if self._process.is_alive():
                self._process.terminate()
            delattr(self, "_process")
        else:
            msg = ("Device {} failed to stop child process {}. Child process is not "
                   "currently running.").format(self.device_name, self.process_name)
            logger.error(msg)
            raise RuntimeError(msg)

    def is_command_done(self):
        """Returns True if command queue is empty, false otherwise.

        Returns:
            bool: True if command queue is empty
        """

        return self._command_queue.empty()

    def _do_work(self):
        """Perform subprocess work and indicate if process should continue.

        Returns:
            bool: indicating process should continue running

        Note:
            This implementation just returns False.
        """

        return False

    def _post_run_hook(self):
        """Cleanup hook for derived classes after process terminates.

        Note:
            Try to avoid raising any exceptions from this method as any
            exception raised will not reach the main process.
        """

        pass

    def _pre_run_hook(self):
        """Setup hook for derived classes before child process run loop begins.

        Returns:
            bool: A boolean value indicating the process loop should proceed.

        Note:
            Returning None or False from this method will prevent the process
            loop from proceeding.
        """

        return True



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class for SSH devices."""
import time

from gazoo_device import config
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import gazoo_device_base
from gazoo_device.capabilities import file_transfer_scp
from gazoo_device.capabilities import ethernet_switch_api_default
from gazoo_device.capabilities import shell_ssh
from gazoo_device.utility import host_utils

logger = gdm_logger.get_gdm_logger()

INFO_PREFIX = "INFO_"
COMMANDS = {
    "BOOT_UP_COMPLETE": "echo 'gdm hello'",
    "LOGGING": "tail -F -n /var/log/messages",
    "GDM_HELLO": "echo 'gdm hello'"
}
REGEXES = {}
TIMEOUTS = {
    "BOOT_UP": 60,
    "DISCONNECT": 60,
    "PING_TO_SSH_DELAY": 10,
    "SHELL": 10,
    "SHELL_DEVICE_RESPONSIVENESS": 5
}


class SshDevice(gazoo_device_base.GazooDeviceBase):
    """Base class for SSH devices."""
    COMMUNICATION_TYPE = "SshComms"
    _COMMUNICATION_KWARGS = {
        "args": host_utils.DEFAULT_SSH_OPTIONS,
        "log_cmd": COMMANDS["LOGGING"],
        "ssh_key_type": None,
        "username": "root"
    }
    _default_filters = ["linux/basic.json", "linux/crashes.json"]

    def __init__(self,
                 manager,
                 device_config,
                 log_file_name=None,
                 log_directory=None):
        super().__init__(manager,
                         device_config,
                         log_file_name=log_file_name,
                         log_directory=log_directory)
        self._commands.update(COMMANDS)
        self._regexes.update(REGEXES)
        self._timeouts.update(TIMEOUTS)

    @decorators.LogDecorator(logger)
    def after_boot_hook(self):
        """Executed after the device boots up (after reboot, upgrade, factory reset).

        Note: no-op by default. Override in derived classes as necessary.
        """
        pass

    @decorators.health_check
    def check_device_responsiveness(self):
        """Check if the device is responsive on console.

        Raises:
            DeviceNotResponsiveError: if device is not responsive on console.
        """
        cmd, timeout = self.commands["GDM_HELLO"], self.timeouts["SHELL_DEVICE_RESPONSIVENESS"]
        try:
            self.shell(cmd, timeout=timeout)
        except errors.GazooDeviceError as err:
            raise errors.DeviceNotResponsiveError(
                self.name, "unable to execute command {!r} on device's shell".format(cmd),
                timeout=timeout, details=str(err))

    @classmethod
    def convert_build_date(cls, build_date):
        """Hook for build date conversion. No-op by default. Override if necessary."""
        return build_date

    @decorators.CapabilityDecorator(
        ethernet_switch_api_default.EthernetSwitchApiDefault)
    def ethernet_switch_api(self):
        """EthernetSwitchApi capability.

        Returns:
            EthernetSwitchApi: EthernetSwitchApi capability instance.
        """
        return self.lazy_init(
            ethernet_switch_api_default.EthernetSwitchApiDefault,
            device_name=self.name,
            address=self.ethernet_switch_ip,
            port=self.ethernet_switch_port,
            disable_port_method=self.ethernet_switch.turn_off_port,
            enable_port_method=self.ethernet_switch.turn_on_port,
            get_port_status_method=self.ethernet_switch.get_port_status,
            pre_disable_port_method=self._pre_disable_ethernet_connection,
            post_disable_port_method=self._post_disable_ethernet_connection,
            pre_enable_port_method=self._pre_enable_ethernet_connection,
            post_enable_port_method=self._post_enable_ethernet_connection)

    @property
    def ethernet_switch(self):
        """Switch for enabling/disabling the device's ethernet."""
        return getattr(self, "_ethernet_switch", None)

    @ethernet_switch.setter
    def ethernet_switch(self, obj):
        """Sets ethernet_switch property to the provided object.

        Args:
            obj (object): Ethernet switch object.

        Note:
            Assumes switch has methods 'turn_on_port', 'turn_off_port', and 'get_port_status'.
        """
        self._ethernet_switch = obj

    @decorators.SettableProperty
    def ethernet_switch_port(self):
        return self.props["optional"].get("ethernet_switch_port")

    @ethernet_switch_port.setter
    def ethernet_switch_port(self, value):
        """Sets ethernet switch port value."""
        self.props["optional"]["ethernet_switch_port"] = value

    @decorators.SettableProperty
    def ethernet_switch_ip(self):
        return self.props["optional"].get("ethernet_switch_ip")

    @ethernet_switch_ip.setter
    def ethernet_switch_ip(self, value):
        """Sets ethernet switch ip value."""
        self.props["optional"]["ethernet_switch_ip"] = value

    @decorators.CapabilityDecorator(file_transfer_scp.FileTransferScp)
    def file_transfer(self):
        """File transfer capability for moving files from and to the device.

        Returns:
            FileTransferScp: file transfer capability using "scp" command.
        """
        return self.lazy_init(file_transfer_scp.FileTransferScp,
                              ip_address_or_fn=self.ip_address,
                              device_name=self.name,
                              add_log_note_fn=self.switchboard.add_log_note,
                              user=self._COMMUNICATION_KWARGS["username"],
                              ssh_key_type=self._COMMUNICATION_KWARGS["ssh_key_type"])

    @decorators.LogDecorator(logger)
    def get_detection_info(self):
        """Gets the persistent and optional attributes of a device.

        Returns:
            tuple: (persistent properties dict, optional properties dict)
        """
        self._set_optional_props()
        self._set_persistent_properties()
        self._format_persistent_properties()

        # Reset all capabilities which could have been instantiated during get_detection_info
        self.reset_all_capabilities()

        return self.props["persistent_identifiers"], self.props["options"]

    @decorators.PersistentProperty
    def health_checks(self):
        """Returns list of methods to execute as health checks."""
        return [self.device_is_connected,
                self.check_create_switchboard,
                self.check_device_responsiveness]

    @decorators.PersistentProperty
    def ip_address(self):
        return self.communication_address

    @classmethod
    def is_connected(cls, device_config):
        """Checks whether or not the device is connected to the computer.

        Args:
            device_config (dict): contains "persistent" dict

        Returns:
            bool: True if device is pingable, False otherwise.
        """
        ip_address = device_config["persistent"]["console_port_name"]
        return host_utils.is_pingable(ip_address)

    @decorators.LogDecorator(logger)
    def recover(self, error):
        """Attempts to recover device based on the type of error specified.

        Args:
            error(CheckDeviceReadyError): The exception thrown that will be used to
                                          identify a possible recovery solution to use.

        Raises:
            GazooDeviceError: If device recovery fails while attempting to perform recovery steps.
            CheckDeviceReadyError: If there are no recovery steps available for the error argument,
                                   it will be re-raised directly.
        """
        if isinstance(error, errors.DeviceNotResponsiveError):
            logger.info("{} device not responding. Rebooting device.", self.name)
            self.reboot()
        elif isinstance(error, errors.DeviceNotBootupCompleteError):
            logger.info("{} not fully booted up. Waiting for complete bootup.", self.name)
            self.wait_for_bootup_complete()
        else:
            raise error

    def shell(self, command, command_name="shell",
              timeout=None, port=0, searchwindowsize=config.SEARCHWINDOWSIZE,
              include_return_code=False):
        """Sends command and returns response and optionally return code.

        Args:
            command (str): Command to send to the device.
            command_name (str): Identifier for command.
            timeout (float): Time in seconds to wait for device to respond.
            port (int): Which port to send on, 0 or 1.
            searchwindowsize (int): Number of the last bytes to look at
            include_return_code (bool): flag indicating return code should be returned.

        Raises:
            GazooDeviceError: if communication fails.

        Note:
            Can try multiple times as connection can sometimes fail.
            See shell_capability init args for setting the number of retry attempts.

        Returns:
            str: If include_return_code is False return the device response to the command.
            tuple: If include_return_code is True return the device response and return code.
        """
        timeout = timeout or self.timeouts["SHELL"]
        return self.shell_capability.shell(
            command, command_name=command_name, timeout=timeout, port=port,
            searchwindowsize=searchwindowsize, include_return_code=include_return_code)

    @decorators.CapabilityDecorator(shell_ssh.ShellSSH)
    def shell_capability(self):
        return self.lazy_init(
            shell_ssh.ShellSSH,
            send_and_expect=self.send_and_expect,
            device_name=self.name,
            timeout=self.timeouts["SHELL"],
            tries=2)

    @decorators.LogDecorator(logger)
    def wait_for_bootup_complete(self, timeout=None):
        """Wait until the device finishes booting up and is ready for testing.

        Args:
            timeout (int): max time to wait for the device to finish booting up.

        Raises:
            DeviceNotBootupCompleteError: device failed to finish booting up before the timeout.
        """
        timeout = timeout or self.timeouts["BOOT_UP"]
        shell_resp = ""
        is_shell_responsive = False
        err_msg = ("wait_for_bootup_complete failed. "
                   "Device hasn't finished booting in {}s.".format(timeout))

        try:
            shell_resp, return_code = self.shell(self.commands["BOOT_UP_COMPLETE"],
                                                 include_return_code=True,
                                                 timeout=timeout)
            is_shell_responsive = return_code == 0

        except errors.GazooDeviceError as err:
            reason = "shell error: {!r}".format(err)
            raise errors.DeviceNotBootupCompleteError(self.name, err_msg, reason=reason)

        if not is_shell_responsive:
            reason = ("Shell responded with a non-zero return code. "
                      "Response: {}, code: {}.".format(shell_resp, return_code))
            raise errors.DeviceNotBootupCompleteError(self.name, err_msg, reason=reason)

    def _format_persistent_properties(self):
        """Cleans up the format of certain persistent properties."""
        persistent_dict = self.props["persistent_identifiers"]
        for key, value in persistent_dict.items():
            if value:
                if "mac_address" in key:
                    persistent_dict[key] = value.replace(":", "")
                elif "code" in key:
                    persistent_dict[key] = value.splitlines()[-1]
                elif key == "build_date":
                    persistent_dict[key] = self.convert_build_date(value)

    def _pre_enable_ethernet_connection(self):
        """Hook called before Ethernet connection to the device is re-enabled."""
        pass

    def _post_enable_ethernet_connection(self):
        """Hook called after Ethernet connection to the device is re-enabled."""
        self.switchboard.open_all_transports()

    def _pre_disable_ethernet_connection(self):
        """Hook called before Ethernet connection to the device is disabled."""
        self.switchboard.close_all_transports()

    def _post_disable_ethernet_connection(self):
        """Hook called after Ethernet connection to the device is disabled."""
        pass

    def _set_optional_props(self):
        """Sets the optional properties for the device during detection."""
        self.props["options"] = {
            "ethernet_switch_ip": None,
            "ethernet_switch_port": None,
            "user_account": None,
            "user_name": None
        }

    def _set_persistent_properties(self):
        """Obtains persistent properties from the device during detection."""
        for key in [cmd for cmd in list(self.commands) if cmd.startswith(INFO_PREFIX)]:
            prop = key.lower()[len(INFO_PREFIX):]  # Remove INFO_ from key
            response = ""
            resp_list = self.shell(self.commands[key]).split("=")
            if resp_list:
                response = resp_list[-1]

            self.props["persistent_identifiers"][prop] = response

    def _verify_reboot(self, disconnect_timeout=None, bootup_timeout=None):
        """Verifies the device goes offline, comes back online, and finishes booting up.

        Args:
            disconnect_timeout (int): max time to wait for device to go offline & come back online.
            bootup_timeout (int): max time to wait for device to finish booting up.

        Raises:
            DeviceNotBootupCompleteError: device failed to boot up.
        """
        disconnect_timeout = disconnect_timeout or self.timeouts["DISCONNECT"]
        bootup_timeout = bootup_timeout or self.timeouts["BOOT_UP"]

        start_time = time.time()
        max_disconnect_time = start_time + disconnect_timeout

        last_ping = True
        while time.time() < max_disconnect_time:
            ping = host_utils.is_pingable(self.ip_address)
            if not ping and not last_ping:
                logger.info("{} offline in {}s.".format(self.name, int(time.time() - start_time)))
                break
            time.sleep(.5)
            last_ping = ping
        else:
            raise errors.DeviceNotBootupCompleteError(
                self.name, "failed to go offline in {}s".format(disconnect_timeout))

        # close ssh transport as the ssh connection is disconnected.
        self.switchboard.close_all_transports()

        time.sleep(5)  # wait for a bit
        start_time = time.time()
        while time.time() < max_disconnect_time:
            if host_utils.is_pingable(self.ip_address):
                logger.info("{} online in {}s".format(self.name, int(time.time() - start_time)))
                break
            time.sleep(.5)
        else:
            raise errors.DeviceNotBootupCompleteError(
                self.name,
                "failed to come online {}s".format(int(time.time() - start_time)))

        # There's a delay between the device being responsive to ping
        # and being able to open SSH connections
        time.sleep(self.timeouts["PING_TO_SSH_DELAY"])

        # Reopen ssh transport as ssh connection is once more available.
        self.switchboard.open_all_transports()

        self.wait_for_bootup_complete(timeout=bootup_timeout)
        logger.info("{} booted up successfully in {}s.".format(
            self.name, int(time.time() - start_time)))

        self.after_boot_hook()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testbed module."""
from __future__ import absolute_import
import subprocess
import sys
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.utility import host_utils

logger = gdm_logger.get_gdm_logger()

COMMANDS = {
    "FILE_SYSTEMS": "df",
    "WIFI_SCAN": "nmcli -t -f SSID device wifi",
    "WIFI_SCAN_MAC": "airport -s"
}


class Testbed(object):
    """Class for testbed interaction."""

    def __init__(self, devices, testing_props, name="testbed"):
        """Initialize a Testbed object.

        Args:
            devices (list): list of device instances.
            testing_props (dict): properties of the testbed used for testing.
            name (str): optional name of the testbed, defaulted to "testbed".
        """
        self.name = name
        self.devices = devices
        self.testing_props = testing_props
        self.commands = COMMANDS

        # map from testing prop to corresponding testbed health check method
        # e.g. if { "wifi_ssid": self.check_wifi_ssid_available }, then the
        #      check_wifi_ssid_available method will be executed if the "wifi_ssid" prop is
        #      provided during Testbed initialization.
        self._PROP_TO_HEALTH_CHECK = {
            "external_storage": self.check_external_storage_mounted,
            "wifi_ssid": self.check_wifi_ssid_available
        }

    @decorators.health_check
    def check_external_storage_mounted(self, enabled):
        """Check if external USB storage is mounted on devices in the testbed.

        Args:
            enabled (str): whether to run this health check, "true" or "false".

        Raises:
            TestbedMissingExternalStorageError: if devices are missing external USB storage.
        """
        if enabled.lower() == "false":
            return

        file_system_cmd = self.commands["FILE_SYSTEMS"]
        usb_storage_location = "/data/usb"
        devices_missing_external_storage = []
        for device in self.devices:
            file_system_output = device.shell(file_system_cmd)
            file_system_mounts = self._get_filesystem_mounts(file_system_output)
            if usb_storage_location not in file_system_mounts:
                devices_missing_external_storage.append(device.name)

        if devices_missing_external_storage:
            raise errors.TestbedMissingExternalStorageError(
                device_names=devices_missing_external_storage,
                details=(f"command {file_system_cmd!r} did not display any filesystems "
                         f"mounted on {usb_storage_location!r}"),
                recovery="mount external USB storage")

    @decorators.health_check
    def check_wifi_ssid_available(self, wifi_ssid):
        """Check if a wifi network is available given a service set identifier (SSID).

        Args:
            wifi_ssid (str): SSID of wifi network.

        Raises:
            TestbedWifiNotAvailableError: if wifi network is not available.

        """
        available_wifi_networks = self._get_available_wifi_networks()
        if wifi_ssid not in available_wifi_networks:
            raise errors.TestbedWifiNotAvailableError(wifi_ssid, available_wifi_networks)

    @decorators.LogDecorator(logger)
    def check_testbed_ready(self):
        """Checks if the testbed is healthy by executing a series of health check methods."""
        health_check_call = []  # list of tuples: (health check method, prop value)
        for prop in self.testing_props:
            if prop in self._PROP_TO_HEALTH_CHECK:
                health_check_call.append((self._PROP_TO_HEALTH_CHECK[prop],
                                          self.testing_props[prop]))
        self._execute_health_check_methods(health_check_call)

    @decorators.LogDecorator(logger)
    def make_testbed_ready(self, setting="on"):
        """Check testbed readiness and attempt recovery if allowed.

        Args:
            setting (str): 'on'|'off'|'check_only'.

        Raises:
            GazooDeviceError: if testbed health checks fail.
        """
        if setting not in ["on", "off", "check_only"]:
            raise errors.GazooDeviceError(
                "{} failed make_testbed_ready. Setting '{}' is invalid.".format(
                    self.name, setting))
        elif setting == "off":
            return

        # check testbed health
        try:
            self.check_testbed_ready()
        except errors.GazooDeviceError as err:
            if setting == "check_only":
                logger.info("{} skipping recovery".format(self.name))
                raise

            logger.info("{} failed check_testbed_ready with {}".format(self.name, repr(err)))

            # attempt to recover the testbed
            self.recover(err)
            logger.info("{} re-checking readiness after recovery attempt".format(self.name))
            self.check_testbed_ready()
            logger.info("{} successfully recovered to ready state".format(self.name))

    @decorators.LogDecorator(logger)
    def recover(self, error):
        """Attempts to recover testbed based on the type of error specified.

        Args:
            error (CheckTestbedReadyError): The exception thrown that will be used to
                                            identify a possible recovery solution to use.

        Raises:
            CheckTestbedReadyError: If recovery fails.
        """
        raise error

    def _execute_health_check_methods(self, health_check_calls):
        """Execute health checks on the testbed.

        Args:
            health_check_calls (list): list of tuples as (health check method, argument)
                                       to execute as health checks.

        Raises:
            CheckTestbedReadyError: if health check fails.

        Notes:
            - The value of the property that triggered the health check is passed to the health
            check method as an argument; therefore, health check methods will accept a single
            argument (unlike device health checks). All property values will be strings.
            - Some health checks will have "true" / "false" argument values. The health check
            should be skipped when the argument is "false".
            - Order of health check methods matter. Health check execution will stop at the first
            failing health check.
        """
        checks_passed = []
        for step, health_check_call in enumerate(health_check_calls):
            health_check_method, arg = health_check_call
            method_name = health_check_method.__name__
            health_check_name = method_name.replace("_", " ").strip().capitalize()

            try:
                health_check_method(arg)
            except errors.CheckTestbedReadyError as err:
                logger.info("testbed health check {}/{} failed: {}.".format(
                    step + 1, len(health_check_calls), health_check_name))
                err.checks_passed = checks_passed
                err.properties = self.testing_props.copy()
                raise

            checks_passed.append("{}.{}".format(type(self).__name__, method_name))
            logger.info("testbed health check {}/{} succeeded: {}.".format(
                step + 1, len(health_check_calls), health_check_name))

    def _get_available_wifi_networks(self):
        """Scan host for available wifi networks.

        Returns:
            list: of available wifi network SSIDs.

        Raises:
            GazooDeviceError: if required nmcli binary does not exist on host.
        """
        command = "WIFI_SCAN_MAC" if "darwin" in sys.platform else "WIFI_SCAN"
        command_args = self.commands[command].split()
        command_name = command_args[0]
        if not host_utils.has_command(command_name):
            raise errors.GazooDeviceError(
                "'{}' command not found and is required to perform wifi scan".format(
                    command_name))

        available_wifi_networks = subprocess.check_output(command_args).decode("utf-8", "replace")

        # wifi scan on mac requires some additional string parsing
        if "darwin" in sys.platform:
            available_network_ssids = []
            ssid_column_end_index = available_wifi_networks.index("SSID") + len("SSID")
            for network_info in available_wifi_networks.splitlines()[1:]:
                if network_info.strip():
                    available_network_ssids.append(network_info[:ssid_column_end_index].strip())
        else:
            available_network_ssids = available_wifi_networks.strip().splitlines()

        return available_network_ssids

    def _get_filesystem_mounts(self, file_system_output):
        """Parse locations that file systems are mounted on from command output.

        Args:
            file_system_output (str): output of command listing device file systems.

        Returns:
            list: locations that file systems are mounted on on the device.
        """
        file_system_mounts = []
        for file_system in file_system_output.splitlines():
            if not file_system:
                continue
            file_system_info = file_system.split()

            # 6 columns of information
            # Filesystem,1K-blocks,Used,Available,Use%,Mounted on
            if len(file_system_info) == 6:
                file_system_mounts.append(file_system_info[5])

        return file_system_mounts



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gazoo Device Manager Fire - CLI using Python Fire and Python API for testing Gazoo devices."""
from __future__ import absolute_import
from __future__ import print_function
import fire
import sys
from gazoo_device import errors
from gazoo_device import fire_manager
from gazoo_device import fire_patch
from gazoo_device import gdm_logger
from gazoo_device import version

logger = gdm_logger.get_gdm_logger()

VERSION_FLAG = "-v"
FLAG_MARKER = "--"
OMIT_FLAGS = ["help"]


def get_flags(args):
    """Parse flags out of array of CLI args.

    Args:
        args (arr): array of arguments provided by the user.

    Returns:
        dict: dictionary of flags to pass to the CLI.

    Note:
        Flags in OMIT_FLAGS dict will not be returned.
    """
    flags = {}
    for arg in args:
        if arg.startswith(FLAG_MARKER):
            flag_name = arg[len(FLAG_MARKER):]
            if flag_name and flag_name not in OMIT_FLAGS:
                flags[flag_name] = True
        else:
            break  # ignore flags after initial CLI call
    return flags


def main(command=None):
    """Main function for Gazoo Device (gazoo_device) package.

    Args:
        command (string): command to pass to Python Fire.

    Returns:
        int: 0 if success, 2 if not.
    """
    retval = 0
    if VERSION_FLAG in sys.argv or (command and VERSION_FLAG in command):
        logger.info('Gazoo Device Manager {}'.format(version))
        return retval

    # parse flags out of commands
    # e.g. "gdm --debug - devices" -> flags = {"debug": True}, commands = ["-", "devices"]
    if command:
        args = command.split()
    else:
        args = sys.argv[1:]
    flags = get_flags(args)
    commands = [arg for arg in args if arg[len(FLAG_MARKER):] not in flags.keys()]

    # instantiate FireManager instance with provided flags
    manager_inst = fire_manager.FireManager(**flags)

    # execute CLI with commands
    try:
        fire_patch.apply_patch()
        fire.Fire(manager_inst, commands, name='gdm')
    except (ValueError, errors.GazooDeviceError) as err:
        logger.error((repr(err)))
        retval = 1
    except KeyboardInterrupt:
        retval = 2
    finally:
        manager_inst.close()

    return retval


if __name__ == '__main__':
    main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TransportBase is the standard GDM transport interface.

is_open(): Returns true if transport is open

open(): Opens transport using properties provided during initialization

close(): Closes transport

read(size=1, timeout=None): Reads up to size bytes from device and
returns them or None if no bytes are read and timeout != None.

write(data, timeout=None): Writes bytes to device and returns the number
of bytes written or None if no bytes were written and timeout != None.

get_property(key, value=None): Returns a transport property identified by
key or if the property doesn't exist yet returns value provided.

get_all_properties(): Returns a deepcopy dictionary of all properties keys
and values currently set.

get_property_list(): Returns the list of property keys

set_property(key, value): Sets a transport property identified by key to
the value provided.
"""
from __future__ import absolute_import
import abc
import copy
from gazoo_device.switchboard import transport_properties
import six


class TransportBase(six.with_metaclass(abc.ABCMeta, object)):
    """Interface for transport classes."""

    def __init__(self, auto_reopen=False, open_on_start=True):
        """Initialize the transport interface.

        Args:
            auto_reopen (bool): flag indicating transport should be reopened if
                                unexpectedly closed.
            open_on_start (bool): flag indicating transport should be open on
                                  TransportProcess start.
        """
        if not hasattr(self, "_properties"):
            self._properties = {}
        self._properties.update({
            transport_properties.AUTO_REOPEN: auto_reopen,
            transport_properties.OPEN_ON_START: open_on_start
        })

    @abc.abstractmethod
    def is_open(self):
        """Returns True if the transport is open.

        Returns:
            bool: True if transport is open, False otherwise.
        """

    @abc.abstractmethod
    def _open(self):
        """Open the transport. Override in derived classes.

        Note: use properties provided during __init__.
        """

    @abc.abstractmethod
    def _close(self):
        """Close the transport. Override in derived classes."""

    @abc.abstractmethod
    def _read(self, size=1, timeout=None):
        """Returns bytes read up to max_bytes within timeout in seconds specified.

        Args:
            size (int): maximum number of bytes to read within timeout seconds
            timeout (float): maximum seconds to wait to read bytes or indefinitely
                             if timeout=None. Default: None

        Returns:
            str: bytes read from transport
                 or "" if no bytes were read
                 or "" if timeout was reached
        """

    @abc.abstractmethod
    def _write(self, data, timeout=None):
        """Writes the data provided within timeout seconds.

        Args:
            data (str): bytes to be written out within timeout seconds
            timeout (float): maximum seconds to wait to write bytes or indefinitely
                             if timeout=None. Default: None

        Returns:
            int: number of bytes written
        """

    def __del__(self):
        """Close the transport on garbage collection."""
        self.close()

    def get_property(self, key, value=None):
        """Returns property matching key specified or value if not set.

        Args:
            key (str): of the property to attempt to retrieve
            value (object): default value to return if the property doesn't exist

        Returns:
            object: The property value for key specified or value if key doesn't exist.
        """
        return self._properties.get(key, value)

    def get_all_properties(self):
        """Returns dictionary of property key-value pairs.

        Returns:
            dict: a deepcopy of all properties as a key value pair
        """
        return copy.deepcopy(self._properties)

    def get_property_list(self):
        """Returns a list of property keys that exist and can be set/retrieved.

        Returns:
            list: of property keys that can be set or retrieved.
        """
        return list(self._properties.keys())

    def set_property(self, key, value):
        """Sets the property specified to the value provided.

        Args:
            key (str): of the property to set
            value (object): of the property to set

        Raises:
            KeyError: when key property doesn't exist
        """
        if key not in self._properties:
            raise KeyError("Property {} doesn't exist for {}.".
                           format(key, self.__class__.__name__))
        self._properties[key] = value

    def open(self):
        """Open the transport if not open already.

        Note: use properties provided during __init__.
        """
        if self.is_open():
            return None

        self._open()

    def close(self):
        """Closes the transport if it's open."""
        if not self.is_open():
            return None

        self._close()

    def read(self, size=1, timeout=None):
        """Returns bytes read up to max_bytes within timeout in seconds specified.

        Args:
            size (int): maximum number of bytes to read within timeout seconds
            timeout (float): maximum seconds to wait to read bytes or indefinitely
                             if timeout=None. Default: None

        Returns:
            str: bytes read from transport
                 or None if transport is NOT open
                 or "" if no bytes were read
                 or "" if timeout was reached

        Note:
            If transport is not opened this method will always return None
        """
        if not self.is_open():
            return None

        return self._read(size, timeout)

    def write(self, data, timeout=None):
        """Writes the data provided within timeout seconds if transport is open.

        Args:
            data (str): bytes to be written out within timeout seconds
            timeout (float): maximum seconds to wait to write bytes or indefinitely
                             if timeout=None. Default: None

        Returns:
            int: number of bytes written or None if no bytes were written
        """
        if not self.is_open():
            return None

        return self._write(data, timeout)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains APIs for interfacing with a Yepkit hub device.

Use the following commands to control the On/Off of the devices connected
to the YKUSH board downstream ports.

Power down a device connected to YKUSH board downstream port 1:

$ ykushcmd -d 1

Power up a device connected to YKUSH board downstream port 1:

$ ykushcmd -u 1

To find out the serial number of a YKUSH board, connect it to the host and
execute the command:

$ ykushcmd -l

Power down a device connected to YKUSH board downstream port 2 of YKB4388:

$ ykushcmd -s YKB4388 -d 2

Power up a device connected to YKUSH board downstream port 2 of YKB4388:

$ ykushcmd -s YKB4388 -u 2

To get the current switching status (UP/DOWN) of a downstream port 2.

$ ykushcmd -g 2
"""
import os
import subprocess

from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.capabilities import switch_power_usb_default
from gazoo_device.utility import host_utils

logger = gdm_logger.get_gdm_logger()

yepkit_enabled = not bool(os.system("which ykushcmd > /dev/null"))

MODEL = "yepkit"
PRODUCT_ID = "f2f7"
VENDOR_ID = "04d8"
YKUSH_PORT_MAP = {"1": 1, "2": 2, "3": 3}

COMMANDS = {
    "POWER_OFF": "ykushcmd -s {} -d {}",
    "POWER_ON": "ykushcmd -s {} -u {}",
    "GET_MODE": "ykushcmd -s {} -g {}",
    "LIST_YEPKITS": "ykushcmd -l"
}

REGEXES = {
    "GET_MODE_REGEX": r"is[ ]([A-Z]+)",
}

TIMEOUTS = {}


class Yepkit(auxiliary_device.AuxiliaryDevice):
    """This class serves as a Python interface to a Yepkit hub."""
    COMMUNICATION_TYPE = "YepkitComms"
    DETECT_MATCH_CRITERIA = {detect_criteria.GENERIC_QUERY.always_true: True}
    DEVICE_TYPE = "yepkit"
    _instances = {}
    _MODEL = MODEL

    def __new__(cls,
                manager,
                device_config,
                log_file_name=None,
                log_directory=None):
        """Object associated with the serial_number (new or previous if already instantiated).

        Args:
            manager (gazoo_device.Manager): manager object for this instance.
            device_config (dict): dict of two dicts, 'optional' device parameters and
                                  'persistent' device parameters.
            log_file_name (str): file name in the log directory for device logs.
            log_directory (str): directory in which the controller will create the log file.

        Returns:
            Yepkit: instance to matching serial port path

        Raises:
            RuntimeError: if yepkit is not enabled.
        """

        if not yepkit_enabled:
            raise RuntimeError("yepkit is not enabled.")
        identifier = device_config['persistent']['console_port_name']
        if identifier not in cls._instances:
            obj = super(Yepkit, cls).__new__(cls)
            cls._instances[identifier] = obj

        return cls._instances[identifier]

    def __init__(self,
                 manager,
                 device_config,
                 log_file_name=None,
                 log_directory=None):
        """Initialize the Yepkit class with the serial port information provided.

        Args:
            manager (gazoo_device.Manager): manager object for this instance of gazoo_device.
            device_config (dict): dict of two dicts, 'optional' device parameters and
                                  'persistent' device parameters.
            log_file_name (str): file name in the log directory for device logs.
            log_directory (str): directory in which the controller will create the log file.

        Raises:
            RuntimeError: if serial number is too short.
        """
        super().__init__(manager,
                         device_config,
                         log_file_name=log_file_name,
                         log_directory=log_directory)
        self._commands.update(COMMANDS)
        self._regexes.update(REGEXES)
        self._timeouts.update(TIMEOUTS)

    @decorators.PersistentProperty
    def health_checks(self):
        """Returns list of methods to execute as health checks."""
        return [self.device_is_connected]

    @decorators.PersistentProperty
    def total_ports(self):
        """Number of ports on Yepkit."""
        return 3

    @decorators.CapabilityDecorator(switch_power_usb_default.SwitchPowerUsbDefault)
    def switch_power(self):
        """Returns a switch_power object to send commands.

        Returns:
            SwitchPowerUsbDefault: switch_power_usb_default
        """
        return self.lazy_init(
            switch_power_usb_default.SwitchPowerUsbDefault,
            shell_fn=self.shell,
            regex_shell_fn=self.shell_with_regex,
            command_dict=self.commands,
            regex_dict=self.regexes,
            device_name=self.name,
            serial_number=self.serial_number,
            total_ports=self.total_ports)

    @decorators.LogDecorator(logger)
    def get_detection_info(self):
        """Gets the persistent and optional attributes of a yepkit.

        Returns:
          tuple: (dict, dict) dictionary of persistent attributes,
                  dictionary of optional attributes (set to None).

        Raises:
            RuntimeError: if yepkit is not enabled.

        Notes:
            persistent: device_type, model, hub_port_name, serial_number,
                        total_ports

          optional: {}
        """
        if not yepkit_enabled:
            raise RuntimeError("yepkit is not enabled.")

        if len(self.communication_address) < 4:
            raise RuntimeError("{} serial number {} is too short. "
                               "Length of serial number must be >= 4".
                               format(self.DEVICE_TYPE, self.serial_number))

        self.props['options'] = {}

        persistent_dict = self.props["persistent_identifiers"]
        persistent_dict["serial_number"] = self.communication_address
        persistent_dict["model"] = self._MODEL
        persistent_dict["total_ports"] = 3

        return self.props["persistent_identifiers"], self.props['options']

    @classmethod
    def is_connected(cls, device_config):
        """Checks whether or not the yepkit is connected to the computer.

        Args:
          device_config (dict): contains "persistent" dict

        Returns:
            bool: whether device serial_number is in host_utils.get_all_yepkit_serials

        Raises:
            RuntimeError: if yepkit is not enabled.
        """
        return device_config["persistent"][
            "console_port_name"] in host_utils.get_all_yepkit_serials()

    @decorators.LogDecorator(logger)
    def recover(self, error):
        """Attempts to recover device based on the type of error specified.

        Args:
            error (CheckDeviceReadyError): A subclass of CheckDeviceReadyError that will be
                used to identify a possible recovery solution to use.

        Raises:
            CheckDeviceReadyError: if no recovery options available for error.
        """
        raise error

    def shell(self,
              command,
              command_name="shell",
              timeout=None,
              port=0,
              include_return_code=False):
        """Sends command and returns response.

        Args:
            command (str): Command to send to the device.
            command_name (str): Optional identifier to use in logs for this command.
            timeout (float): Seconds to wait for pattern after command sent.
            port (int): no op.
            include_return_code (bool): flag indicating return code should be returned

        Returns:
            str or tuple: response or (response, return_code) if include_return_code is True.
                          return_code 0 - success, return_code -1 - error.

        Raises:
            GazooDeviceError: if command execution fails OR
                             couldn't find the requested group in any of the responses.
        """
        timeout = timeout or self.timeouts["SHELL"]
        if "ykushcmd" not in command:
            raise errors.GazooDeviceError("{} shell expected 'ykushcmd' instead found: {!r}".
                                          format(self.name,
                                                 command))
        raw_list = command.split(" ")
        if include_return_code:
            result = subprocess.check_output(raw_list).decode("utf-8", "replace")
            if "Application Help" in result:
                return result, -1  # Error
            return result, 0
        return subprocess.check_output(raw_list).decode("utf-8", "replace")

    def shell_with_regex(self,
                         command,
                         regex,
                         regex_group=1,
                         command_name="shell",
                         raise_error=False,
                         tries=1,
                         port=0):
        """Sends a command, searches for a regex in the response, and returns a match group.

        Args:
            command (str): command to issue.
            regex (str): regular expression with one or more capturing groups.
            regex_group (int): number of regex group to return.
            command_name (str): command name to appear in log messages.
            raise_error (bool): whether or not to raise error if unable to find a match.
            tries (int): how many times to try executing the command before failing.
            port (int): which port to send the shell command to.

        Returns:
            str: value of the capturing group with index 'regex_group' in the match.

        Raises:
            GazooDeviceError: if command execution fails OR
                             couldn't find the requested group in any of the responses.
        """
        return self.command_with_regex(command,
                                       regex,
                                       self.shell,
                                       regex_group=regex_group,
                                       raise_error=raise_error,
                                       tries=tries,
                                       command_name=command_name,
                                       port=port)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""build utils for obtaining firmware builds.

All builds are unzipped to /tmp/<self.name>/<build_package_name>
"""
from __future__ import absolute_import
import json
import os
import re
import shutil
import tarfile
import traceback
import zipfile
from gazoo_device import config
from gazoo_device import gdm_logger
from gazoo_device.utility import host_utils
from gazoo_device.utility import http_utils

logger = gdm_logger.get_gdm_logger()

BUILD_FOLDER = "/tmp/{name}"
GSUTIL_PREFIX = "gs://"

TEMPLATES = {
    "build_url_notimplemented": "NotImplemented",
    "firmware_notimplemented": "NotImplemented",
}

NOT_IMPLEMENTED = {
    "build_url_template": "build_url_notimplemented",
    "firmware_template": "firmware_notimplemented",
    "build_number": None
}

with open(config.DEFAULT_BUILD_INFO_FILE) as default_build_json:
    DEFAULT_BUILD_DICTIONARY = json.loads(default_build_json.read())

    # set default build info for unimplemented devices
    for device in DEFAULT_BUILD_DICTIONARY:
        if not DEFAULT_BUILD_DICTIONARY[device]:
            DEFAULT_BUILD_DICTIONARY[device] = NOT_IMPLEMENTED.copy()


class BuildObtainer(object):
    """Class for obtaining builds from known locations.

    Notes:
        Builds get downloaded to /tmp/<name>
    """

    def __init__(self, name, owner=""):
        """Ininitalize the build obtainer.

        Args:
            name (str): unique identifier for device. Determines where builds get downloaded.
            owner (str): optional LDAP of owner if initialize from a device class.
        """
        self.name = name
        self.owner = owner
        self.local_build_dir = BUILD_FOLDER.format(name=self.name)
        self.close()  # removes previous version of directory
        os.makedirs(self.local_build_dir, mode=0o777)
        self.build_types = {
            GSUTIL_PREFIX: self._obtain_gsutil_build
        }

        self._build_autocompleter = BuildAutoCompleter(owner=owner)
        self.templates = TEMPLATES.copy()

    def close(self):
        """Removes the build folder upon close."""
        try:
            if os.path.isdir(self.local_build_dir):
                shutil.rmtree(self.local_build_dir)
            elif os.path.isfile(self.local_build_dir):
                os.remove(self.local_build_dir)
        except Exception as err:
            logger.warning(
                "{} failed to remove build_dir {}. Err: {}".format(
                    self.name, self.local_build_dir, str(err)))

    def get_version(self, device_type, build_info):
        """Converts build_info to absolute build path and extracts build version.

        Args:
          device_type (str): device type.
          build_info (dict): build information provided by parent method.

        Returns:
          str: version in build path.

        Notes:
           Errors will be raised if files not found.
        """
        path, regex = self._build_autocompleter.get_build_url_and_file_identifier(
            device_type, is_version_check=True, **build_info)
        version = self.get_version_from_folder_files(path, regex)
        logger.info("{} target firmware_version is {}".format(self.name, version))
        return version

    def get_version_and_build_path(self, device_type, build_info):
        """Converts build_info to absolute build path and extracts build version.

        Args:
          device_type (str): device type.
          build_info (dict): build information provided by parent method.

        Returns:
          tuple: version in build path and build_path

        Notes:
           Errors will be raised if files not found.
        """
        path, regex = self._build_autocompleter.get_build_url_and_file_identifier(
            device_type, is_version_check=True, **build_info)
        version = self.get_version_from_folder_files(path, regex)
        logger.info("{} target firmware_version is {}".format(self.name, version))
        return version, path

    def get_firmware_type(self, device_type, build_info):
        """Returns the target firmware type for the build, i.e. eng or user.

        Args:
            device_type (str): device type.
            build_info (dict): build information provided by parent method.

        Returns:
            str: firmware type or "UNKNOWN" if unable to extract type from build_info.
        """
        build_type = "UNKNOWN"
        build_type = self._build_autocompleter.get_default_values(device_type).get(
            "build_type", build_type)

        if "build_type" in build_info:
            build_type = build_info["build_type"]
        elif "build_url" in build_info:
            build_type_from_url = self._get_build_type_from_build_url(
                build_info["build_url"], device_type)
            if build_type_from_url != "UNKNOWN":
                build_type = build_type_from_url

        if build_type != "UNKNOWN":
            logger.info("{} target firmware_type is {}".format(self.name, build_type))
        return build_type

    def locate_and_obtain_build(self, device_type, build_info=None):
        """Converts build_info to absolute build path and extracts build file to local path.

        Args:
          device_type (str): device type.
          build_info (dict): build information provided by parent method.

        Returns:
          str: local path to file.

        Raises:
           RuntimeError: if no files found.

        Notes:
           Errors will be raised if files not found.
        """
        if not build_info:
            build_info = {}
        path, regex = self._build_autocompleter.get_build_url_and_file_identifier(
            device_type, **build_info)
        build_info.update({"build_url": path})  # for obtaining payload properties if necessary
        files = self.get_files_matching_pattern(path, regex)
        if not files:
            raise RuntimeError("No files at {} matching {!r}".format(path, regex))
        firmware_file = files[0]
        logger.info("{} obtaining {} from {}".format(self.name, firmware_file, path))
        local_path = self.obtain_build(os.path.join(path, firmware_file))
        return local_path

    def extract_build(self, compressed_file_path, compression_type="tgz"):
        """Extract .tgz build and save it to /tmp/<self.name>/<file_name>/.

        Args:
            compressed_file_path (str): absolute path to tar file.
            compression_type (str): "tar", "tgz" or "zip"

        Returns:
            str: location of extracted build.

        Raises:
            ValueError: if compressed file does not exist.
        Note:
           Removes the compressed file
        """
        if not os.path.exists(compressed_file_path):
            raise ValueError(
                "Unable to extract. File {} doesn't exist".format(compressed_file_path))
        basename = os.path.basename(compressed_file_path)
        folder = os.path.splitext(basename)[0]
        local_folder_path = os.path.join(self.local_build_dir, folder)
        logger.info(
            "{} extracting compressed file {} into {}.",
            self.name,
            folder,
            local_folder_path)
        if compression_type == "zip":
            compressor = zipfile.ZipFile(compressed_file_path, 'r')
        elif compression_type == "tar" or compression_type == "tgz":
            compressor = tarfile.open(compressed_file_path)
        else:
            raise ValueError("Compression type {} is not supported. "
                             "It has to be either tar, tgz or zip".format(compression_type))

        compressor.extractall(path=local_folder_path)
        compressor.close()

        os.remove(compressed_file_path)
        return local_folder_path

    def extract_tgz_build(self, compressed_file_path):
        """Extract .tgz build and save it to /tmp/<self.name>/<file_name>/.

        Args:
            compressed_file_path (str): absolute path to tar file.

        Returns:
            str: location of extracted build.

        Raises:
            ValueError: if compressed file does not exist.
        Note:
           Removes the compressed file
        """
        return self.extract_build(compressed_file_path, "tgz")

    def extract_zip_build(self, compressed_file_path):
        """Extract .tgz build and save it to /tmp/<self.name>/<file_name>/.

        Args:
            compressed_file_path (str): absolute path to zip file.

        Returns:
            str: location of extracted build.

        Raises:
            ValueError: if compressed file does not exist.
        Note:
           Removes the compressed file
        """
        return self.extract_build(compressed_file_path, "zip")

    def get_file_list(self, path):
        """Get files at path.

        Args:
            path (str): absolute path to folder in which files reside(url or gsutil)

        Returns:
            list: list of file names relative to give path.

        Notes:
          file list for url includes all the hyperlinks to files on that page.
          It removes .css files but other oddities might be there.
        """
        path = os.path.join(path, "")
        prefix = self._get_valid_build_prefix(path)
        if prefix == GSUTIL_PREFIX:
            return self._get_file_list_from_gsutil(path)

    def get_version_from_folder_files(self, path, pattern):
        """Finds a file that matches the pattern and returns the version from it.

        Args:
            path (str): build path at which to find files
            pattern (regex): pattern with one group through through to extract version from a file.

        Raises:
            ValueError: if no match

        Returns:
            str: matching data

        Note:
            Cameras do not have the full version in their firmware file
            so this method uses a bit of other data.
        """
        files = self.get_file_list(path)
        versions = [re.match(pattern, a_file).group(1)
                    for a_file in files
                    if re.match(pattern, a_file)]
        if not versions:
            raise ValueError("Unable to find {} in {}".format(pattern, files))
        version = versions[0]
        return version

    def get_files_matching_pattern(self, path, pattern):
        """Get files at path that match a pattern.

        Args:
            path(str): absolute path to folder in which files reside(url or gsutil)
            pattern(regex): regex pattern that string should match.

        Returns:
            list: list of file base names(not absolute paths)
        """
        files = self.get_file_list(path)
        return [a_file for a_file in files if re.match(pattern, a_file)]

    def obtain_build(self, build_url):
        """Transfer builds to local path. Uncompresses them if desired.

        Args:
            build_url(str): absolute path to file needed to be obtained.

        Returns:
            str: path of local file
        """
        prefix = self._get_valid_build_prefix(build_url)
        method = self.build_types[prefix]
        return method(build_url)

    def _get_build_type_from_build_url(self, build_url, device_type):
        """Parses the target build type out of a build url.

        Args:
          build_url (str): url to build_folder if it exists.
          device_type (str): type of device associated with build_url.

        Returns:
           str: build_type if detected, else "UNKNOWN".

        Note:
            This method will find a build_type within a build_url by using regex matching against
            the device's build_url template. If the device does not have a default build_url
            or the default build_url does not include a build_type, "UNKNOWN" will be returned.
        """
        if device_type not in self._build_autocompleter.device_build_defaults:
            return "UNKNOWN"

        device_build_defaults = self._build_autocompleter.device_build_defaults[device_type]
        build_url_template_name = device_build_defaults.get("build_url_template")
        build_url_template = self._build_autocompleter.templates[build_url_template_name]
        if "build_type" in build_url_template:

            # replace variables in build url template with regexes
            build_info = device_build_defaults.copy()
            build_info = {key: r"(.*)" for key in build_info}
            build_info["build_type"] = r"(?P<build_type>.*)"
            build_url_regex = build_url_template.format(**build_info)

            # attempt to parse build_type out of regexed build url
            build_type_match = re.match(build_url_regex, build_url)
            if build_type_match:
                return build_type_match.group("build_type")

        return "UNKNOWN"

    def _get_file_list_from_gsutil(self, gsutil_path):
        """Returns list of files / folder at gsutil provided.

        Args:
            gsutil_path (str): Location to use to obtain list of files.

        Returns:
            list: list of relative files / folders at that location.

        Raises:
            RuntimeError: if path does not exist.

        Note:
           'gsutil ls gs://example' output is ['gs://example/file1', 'gs://example/file2'].
           Absolute path is removed to match url format and enable ease of matching file patterns.
        """
        gsutil_path = self._validate_path_or_symlinked_path_exists(gsutil_path)
        output = host_utils.gsutil_command("ls", gsutil_path)
        relative_paths = [path.replace(gsutil_path, "") for path in output.splitlines()]
        return relative_paths

    def _get_file_list_from_jenkins(self, url):
        """Returns list of files at URL provided.

        Note:
            Definition of file is a hyperlink with a "." in it that's not a .css file.
            If server response code from urlopen() is not 200 nor 404,
            the method will retry the call to urlopen() up to MAX_RETRY_URl_OPEN times
            before raising the GazooDeviceError.

        Args:
            url(str): A URL string to use to obtain a list of files.

        Raises:
          RuntimeError: Response to URL is not HTTP code 200.

        Returns:
          list: list of links to files and folders at that location.
        """
        url = os.path.join(url, "api/xml?tree=artifacts%5BrelativePath%5D")
        raw_page = http_utils.read_raw_html_page(url)
        links = re.findall(
            r'<relativePath>([a-z0-9\_\-\./]+)</relativePath>', raw_page)
        return [os.path.join('artifact', link) for link in links]

    def _get_file_list_from_url(self, url):
        """Returns list of files at URL provided.

        Note:
            Definition of file is a hyperlink with a "." in it that's not a .css file.
            If server response code from urlopen() is not 200 nor 404,
            the method will retry the call to urlopen() up to MAX_RETRY_URl_OPEN times
            before raising the GazooDeviceError.

        Args:
            url(str): A URL string to use to obtain a list of files.

        Raises:
          RuntimeError: Response to URL is not HTTP code 200.

        Returns:
          list: list of links to files and folders at that location.
        """
        raw_info = http_utils.read_raw_html_page(url)
        files = re.findall(r'href=\"([\w\/\-\_\.\*]+)\"', raw_info)
        return [a_file for a_file in files
                if not a_file.endswith(".css") and "." in a_file]

    def _get_valid_build_prefix(self, path):
        """Validates and returns build prefix.

        Args:
            path (str): build path.

        Returns:
            str: prefix

        Raises:
            ValueError: if build type is not supported.
        """
        prefix_list = [key for key in self.build_types.keys() if key in path]
        if not prefix_list:
            raise ValueError("Obtaining builds from {} not supported. "
                             "Can only load builds from: {}".format(
                                 path, " and ".join(list(self.build_types.keys()))))
        return prefix_list[0]

    def _obtain_gsutil_build(self, gs_path):
        """Obtains file through the gsutil utility.

        Args:
          gs_path (str): gsutil path starting with prefix.

        Returns:
            str: local file path

        Raises:
            RuntimeError: if gsutil folder isn't accessible by user or build doesn't exist.
        """
        folder = os.path.dirname(gs_path) + "/"
        file_name = os.path.basename(gs_path)
        folder = self._validate_path_or_symlinked_path_exists(folder)
        full_path = os.path.join(folder, file_name)
        logger.info("{} acquiring build from {}".format(self.name, gs_path))
        host_utils.gsutil_command("cp", full_path, [self.local_build_dir])

        local_file_path = os.path.join(self.local_build_dir, file_name)
        if not os.path.exists(local_file_path):
            raise RuntimeError(
                "Unable to copy file from {} to {} via gsutil.".format(full_path, local_file_path))
        logger.info("{} successfully copied build to {}".format(self.name, local_file_path))
        return local_file_path

    def _obtain_url_build(self, url):
        """Obtains file from url and saves it into /tmp/<self.name>.

        Args:
            url(str): url starting with prefix.

        Returns:
            str: local file path

        Raises:
            RuntimeError: if prefix is not accessible or build doesn't exist.
        """
        # Retrieve list of files and download image file found
        file_name = os.path.basename(url)
        local_path = os.path.join(self.local_build_dir, file_name)
        # Check if main url location is accessible
        http_utils.download_file(url, local_path)
        return local_path

    def obtain_package(self, gsutil_path, destination_folder, destination_file):
        """Obtains a package file using gsutil and saves it locally.

        Args:
            gsutil_path (str): gsutil path to the package file.
            destination_folder (str): absolute path to folder to store file locally.
            destination_file (str): filename to use when saving file.

        Returns:
            str: absolute path to downloaded local package file.

        Raises:
            RuntimeError: if file doesn't exist where it should locally.

        Note:
            Packages will be saved to /tmp/<device_name>/packages/<package_name>.
        """
        logger.info("{} acquiring packages from {}.".format(self.name, gsutil_path))
        if not os.path.exists(destination_folder):
            os.makedirs(destination_folder, mode=0o777)
        destination_path = os.path.join(destination_folder, destination_file)

        # download zip file of packages using gsutil
        host_utils.gsutil_command("cp", gsutil_path, [destination_path])
        if not os.path.exists(destination_path):
            raise RuntimeError("Unable to copy file from {} to {} using gsutil.".format(
                gsutil_path, destination_path))

        logger.info("{} successfully copied packages to {}.".format(self.name, destination_path))
        return destination_path

    def _validate_path_or_symlinked_path_exists(self, gsutil_folder):
        """Validates path exists. Find first prefix that is symlink or doesn't exist.

        Args:
           gsutil_folder (str): folder

        Returns:
          str: folder or actual path if path had symlink in it.

        Raises:
          RuntimeError: if path doesn't exist.

        Note:
           Only handles symlinks in images. If the 'images' prefix breaks, this breaks.
        """
        try:
            output = host_utils.gsutil_command("ls", gsutil_folder)

            # First line is always the folder or symlink's name itself.
            if len(output.splitlines()) > 1:
                return gsutil_folder

        except RuntimeError:
            raise RuntimeError("File {} doesn't exist. Error: {}\n"
                               .format(gsutil_folder, traceback.format_exc()))
        return gsutil_folder


class BuildAutoCompleter(object):
    """Object for converting build information to build_url and firmware_identifier.
    """

    def __init__(self, owner=""):
        """Ininitalize the build autocompleter.

        Args:
            owner (str): optional LDAP of owner if initialized from a device class.
        """
        self.owner = owner
        self.templates = TEMPLATES.copy()
        self.device_build_defaults = DEFAULT_BUILD_DICTIONARY.copy()

    def get_build_url_and_file_identifier(self, device_type, is_version_check=False,
                                          build_url=None, **user_info):
        """Returns build path and file_identifier.

        Args:
            device_type (str): device type
            is_version_check (bool): Whether this is a version check.
            build_url (str): path to build folder that holds firmware.
            **user_info (dict): optional dictionary of build info to override default values.

        Return:
            tuple: build_url, file_identifier

        Raises:
            NotImplementedError: not implemented.
        """
        raise NotImplementedError("Not implemented.")

    def get_default_values(self, device_type):
        """Return default_build info for that device type.

        Args:
            device_type (str): device type to get.

        Returns:
            dict: dictionary of build_info
        """
        if device_type in self.device_build_defaults:
            default_build_info = self.device_build_defaults[device_type]

            # autocomplete default templates
            completed_build_templates = {}
            for build_prop_name, build_prop in default_build_info.items():
                if build_prop in self.templates:
                    prop_name = build_prop_name.replace("_template", "_default")
                    try:
                        prop_value = self.templates[build_prop].format(
                            device_type=device_type, **default_build_info)
                    except KeyError:
                        continue
                    completed_build_templates[prop_name] = prop_value
            self.device_build_defaults[device_type].update(completed_build_templates)

            return self.device_build_defaults[device_type]
        logger.warning(
            "{} does not currently support autocompletion of build info.".format(device_type))
        return {}

    def get_supported_device_types(self):
        """Returns all the devices supported by autocompleter.

        Returns:
            list: list of all device_types supported
        """
        device_types = []
        keys = list(self.device_build_defaults.keys())
        for key in keys:
            device_types += key.split("/")
        return device_types



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for usb information."""
from __future__ import absolute_import
import re
import sys
from gazoo_device.utility import usb_config
from gazoo_device.utility import usb_info_linux
from gazoo_device.utility import usb_info_mac


def find_matching_connections(match_criteria):
    """Returns usb_info instances for connections that match criteria.

    Args:
        match_criteria (dict): dictionary of match criteria.

    Returns:
        list: list of UsbInfo instances that match criteria.

    Raises:
        ValueError: if match criteria contains bad key(s).

    Note:
      Connections first excluded and then included.
      If there is no include regex, all instances that dont match exclude_regex are included.

      Dictionary entry: {"<key>": {
                            "include_regex": "<regex>",
                            "exclude_regex": "<regex>""}
                        }
    """
    instances = get_address_to_usb_info_dict().values()
    allowed_keys = usb_config.UsbInfo.get_properties()
    bad_keys = [key for key in match_criteria.keys() if key not in allowed_keys]
    if bad_keys:
        raise ValueError("The keys {} are not allowed. Possible keys: {}".format
                         (bad_keys, allowed_keys))
    for attr, entry in match_criteria.items():
        if "exclude_regex" in entry:
            instances = [instance for instance in instances
                         if not re.search(entry["exclude_regex"], str(getattr(instance, attr)))]
        if "include_regex" in entry:
            instances = [instance for instance in instances
                         if re.search(entry["include_regex"], str(getattr(instance, attr)))]
    return instances


def get_address_to_usb_info_dict():
    """Gets a dictionary of usb devices with all relevent information."""
    if sys.platform == 'darwin':
        module = usb_info_mac
    else:
        module = usb_info_linux
    return module.get_address_to_usb_info_dict()


def get_all_serial_connections():
    """Returns a list of all serial connections."""
    usb_info = get_address_to_usb_info_dict()
    return [key for key, entry in usb_info.items()
            if entry.product_name not in usb_config.ANDROID_NAMES]


def get_device_info(address):
    """Gets the usb info of a specific device.

    Args:
        address (str): serial path or adb serial.

    Returns:
        object: usb_info instance encoding information for that specific address.

    """
    address_to_info_dict = get_address_to_usb_info_dict()
    if address in address_to_info_dict:
        return address_to_info_dict[address]

    return usb_config.UsbInfo()


def get_vendor_number_from_path(address):
    """Returns the vendor number for the address provided.

    Args:
        address (str): address to find matching vendor id for.

    Raises:
        ValueError: if address not found.

    Returns:
        str: vendor number (e.g. "4012") from address.
    """
    usb_info_inst = get_device_info(address)
    return usb_info_inst.vendor_id


def get_ftdi_interface_from_path(address):
    """Returns the FTDI interface embedded in address provided.

    Args:
        address (str): Address to extract interface from.

    Raises:
        ValueError: if FTDI interface can't be obtained from address.

    Returns:
        str: Interface (e.g. "if02") from address.
    """
    usb_info_inst = get_device_info(address)
    return usb_info_inst.ftdi_interface


def get_product_name_from_path(address):
    """Returns the product name for address provided.

    Args:
        address (str): Address to extract FTDI name from.

    Raises:
        ValueError: if product name can't be obtained from address.

    Returns:
        str: product name from address.
    """
    usb_info_inst = get_device_info(address)
    return usb_info_inst.product_name


def get_other_ftdi_line(address, ftdi_interface):
    """Gets a path to the corresponding ftdi line for that address.

    Args:
       address (str): address for an ftdi based serial.
       ftdi_interface (int): number of the ftdi interface to switch to.

    Returns:
       str: address for other ftdi line.

    Raises:
        ValueError: if a matching ftdi line is not found.
    """
    serial_number = get_serial_number_from_path(address)
    devices = find_matching_connections(
        {"serial_number": {"include_regex": serial_number},
         "ftdi_interface": {"include_regex": str(ftdi_interface)}})
    if devices:
        return devices[0].address
    raise ValueError(
        "No connection found with {} and ftdi interface {}".format(
            serial_number, ftdi_interface))


def get_serial_number_from_path(address):
    """Returns the serial number embedded in address provided.

    Args:
        address (str): Address to extract serial number from.

    Raises:
        ValueError: if FTDI serial number can't be obtained from address.

    Returns:
        str: Serial number from address.
    """
    usb_info_inst = get_device_info(address)
    return usb_info_inst.serial_number


def get_usb_hub_info(device_address):
    """Gets the usb hub information for each communication address of the device.

    Args:
        device_address (str): communication address for device.

    Returns:
       dictionary: usb hub name and port number of the device usb connection.
    """
    usb_info = {}
    usb_info["device_usb_hub_name"] = get_usb_hub_address_from_address(device_address)
    usb_info["device_usb_port"] = get_usb_hub_port_from_address(device_address)
    return usb_info


def get_usb_hub_address_from_address(address):
    """Returns the communication address for usb hub the device is connected to.

    Args:
        address (str): communication address of device.

    Returns:
        str: communication address of the usb hub.
        None: if no associated usb hub.
    """
    usb_info_inst = get_device_info(address)
    return usb_info_inst.usb_hub_address


def get_usb_hub_address_from_serial_number(serial_number):
    """Returns the communication address for usb hub the device is connected to.

    Args:
        serial_number (str): device serial_number to associate with port number.

    Returns:
        str: communication address of the usb hub.
        None: if no associated usb hub.
    """
    devices = find_matching_connections(
        {"serial_number": {"include_regex": serial_number}})
    if devices:
        return devices[0].usb_hub_address
    return None


def get_usb_hub_port_from_address(address):
    """Returns the usb hub port number associated with an address.

    Args:
        address (str): communication Address to associate with port number.

    Returns:
        int: Port number of the usb hub the usb address is attached to.
        None: if no associated port.
    """
    usb_info_inst = get_device_info(address)
    return usb_info_inst.usb_hub_port


def get_usb_hub_port_from_serial_number(serial_number):
    """Returns the usb hub port number associated with an address.

    Args:
        serial_number (str): device serial_number to associate with port number.

    Returns:
        int: Port number of the usb hub the device is attached to.
        None: if no associated port.
    """
    devices = find_matching_connections(
        {"serial_number": {"include_regex": serial_number}})
    if devices:
        return devices[0].usb_hub_port
    return None



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Capability base class. All capabilities must inherit from this class.

Serves two purposes:
- Enforce all capabilities to have a device_name attribute.
  This is a requirement for informative log and error messages in multi-device testing.
- Dynamically identify all capabilities supported by GDM.
  This is accomplished by scanning the capabilities folder and identifying all classes
  which subclass CapabilityBase.
"""
from __future__ import absolute_import
import abc
import six
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.utility import common_utils
from gazoo_device.utility import supported_classes

logger = gdm_logger.get_gdm_logger()
_CAPABILITY_INTERFACE_SUFFIX = "_base"
_capability_name_cache = {}


class CapabilityBase(six.with_metaclass(abc.ABCMeta), object):
    """Abstract base class for all capabilities."""

    def __init__(self, device_name):
        """Set the device_name attribute of the capability.

        Args:
            device_name (str): name of the device instance the capability is attached to.

        Note:
            All capabilities need to include the device name attribute to support error and log
            messages in multi-device testing.
        """
        self._device_name = device_name
        self._healthy = False

    @property
    def healthy(self):
        """Whether or not the capability has passed a health_check."""
        return self._healthy

    @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
    def health_check(self):
        """Overridden. Checks capability readiness."""
        self._healthy = True

    @classmethod
    def get_capability_interface(cls):
        """Returns the capability interface class of the given capability flavor or interface.

        Returns:
            type: capability interface class of this capability flavor.

        Note:
            If called on a flavor class, the nearest parent interface will be returned.
                KeypadDefault.get_capability_interface() -> <class keypad_base.KeypadBase>
            If called on an interface class, the same interface class will be returned.
                KeypadBase.get_capability_interface() -> <class keypad_base.KeypadBase>
            If called on CapabilityBase class, returns CapabilityBase (special case).

        Raises:
            GazooDeviceError: if the capability flavor does not inherit from any capability
                             interface.
        """
        if cls is CapabilityBase:
            return CapabilityBase
        for parent_class in cls.__mro__:
            if parent_class in supported_classes.SUPPORTED_CAPABILITY_INTERFACES.values():
                return parent_class
        raise errors.GazooDeviceError(
            "Capability {} does not inherit from a capability interface.".format(cls))

    @classmethod
    def get_capability_name(cls):
        """Generate the name under which a capability is accessible in a device class.

        Returns:
            str: capability name to use in device class capability definitions.

        Note:
            This is the default implementation. Override if necessary.

        Raises:
            GazooDeviceError: if unable to generate the name because the capability interface does
                             not follow the standard naming convention (<some_capability>Base).

        Examples:
            KeypadBase.get_capability_name() -> "keypad"
            KeypadDefault.get_capability_name() -> "keypad"
            COEventsBase.get_capability_name() -> "co_events"
        """
        interface = cls.get_capability_interface()
        snake_case_name = common_utils.title_to_snake_case(interface.__name__)
        if not snake_case_name.endswith(_CAPABILITY_INTERFACE_SUFFIX):
            raise errors.GazooDeviceError(
                "Unable to automatically generate the capability name for capability {} "
                "from interface {}. Interface snake case name {} does not end with '{}'"
                .format(cls, interface, snake_case_name, _CAPABILITY_INTERFACE_SUFFIX))
        return snake_case_name[:-len(_CAPABILITY_INTERFACE_SUFFIX)]

    @decorators.CapabilityLogDecorator(logger, level=None)
    def validate_required_keys(self, key_list, dictionary, dictionary_name):
        """Verify that the required keys are present in the provided dictionary.

        Args:
            key_list (list): list of required keys (string) for the provided dictionary.
            dictionary (dict): dictionary to check for all the required keys.
            dictionary_name (str): name (or purpose) of the dictionary.
                                   Only used in the error message. For example, "Commands".

        Raises:
            GazooDeviceError: if a required key is missing from the dictionary.
            GazooDeviceError: if "dictionary" argument is not a dictionary instance.

        Note:
            self._device_name must be set before calling this method (by CapabilityBase.__init__).
        """
        base_msg = "{} failed to create {!r} capability.".format(self._device_name,
                                                                 self.get_capability_name())
        if not isinstance(dictionary, dict):
            extra_msg = " Expected {!r} to be a dict, found {}".format(dictionary_name, dictionary)
            raise errors.GazooDeviceError(base_msg + extra_msg)

        missing_keys = [key for key in key_list if key not in dictionary]
        if missing_keys:
            extra_msg = " Dictionary {!r} is missing the following keys: {}".format(
                dictionary_name, missing_keys)
            raise errors.GazooDeviceError(base_msg + extra_msg)

    @classmethod
    def _get_capability_name(cls):
        """Add error checking and caching to capability name generation.

        Returns:
            str: capability name to use in device class capability definitions.

        Raises:
            GazooDeviceError: if cls is CapabilityBase.
        """
        if cls is CapabilityBase:
            raise errors.GazooDeviceError(
                "Cannot generate a capability name for CapabilityBase interface.")

        if cls not in _capability_name_cache:
            _capability_name_cache[cls] = cls.get_capability_name()
        return _capability_name_cache[cls]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class to press buttons by controlling GPIO lines on FTDI-based boards."""
from __future__ import absolute_import
import pylibftdi
import time


class FtdiButtons(object):
    """Allows for triggering GPIO lines on FTDI based boards."""

    def __init__(self, ftdi_serial_number, button_map, polarity=False):
        """Initializes the FtdiButtons with the properties provided.

        The ftdi_serial_number is needed to unique identify the FTDI device
        to control as there may be more than one FTDI device available. The
        button map defines the interfaces and pin (bit) to set for each button
        and should look like this:
        some_button_map = {"reset_btn": {"interface": 1, "pin": 5},
                           "button_btn": {"interface": 2, "pin": 9}}

        Args:
            ftdi_serial_number (str): FTDI serial number to use
            button_map (dict): button to pin (bit) mapping.
            polarity (bool):  GPIO pin polarity to use (True to push button).
        """
        self._button_map = button_map
        self._button_down = {}
        self._ftdi_serial_number = ftdi_serial_number
        self._polarity = polarity
        self._sub_process = None
        self._bb_map = {}
        self._bitbang_device_init = False
        self.bitbang_device_init()

    def __del__(self):
        self.close()

    def close(self):
        """Release any held buttons on close."""
        for button in list(self._button_down):
            self.release(button)

        for interface in self._bb_map:
            ftdi_device = self._bb_map[interface]
            ftdi_device.close()

        self._bb_map.clear()
        self._bitbang_device_init = False

    def is_valid(self, button):
        """Return True if button specified is in the list of valid buttons.

        Args:
            button (str): button identifier.

        Returns:
            bool: True if button specified is in button map provided
        """

        return button in list(self._button_map.keys())

    def bitbang_device_init(self):
        """Create BitBangDevice's for each button interface and initiate to 'off'."""
        for button in list(self._button_map.keys()):
            interface = self._button_map[button]["interface"]
            pin = self._button_map[button]["pin"]

            if interface not in self._bb_map:
                ftdi_device = pylibftdi.BitBangDevice(device_id=self._ftdi_serial_number,
                                                      interface_select=interface,
                                                      direction=(1 << pin))
                ftdi_device.port = 0 if self._polarity else 255
                self._bb_map[interface] = ftdi_device
            else:
                self._bb_map[interface].direction |= (1 << pin)
            self._button_down[button] = False

        self._bitbang_device_init = True

    def press(self, button, wait=0.0):
        """Presses button and waits for the time specified.

        Note: You must call release or close to release the button later.

        Args:
            button (str): identify which button to press
            wait (float): seconds to wait before returning

        Raises:
            ValueError: invalid button or wait given.
        """
        if not self._bitbang_device_init:
            self.bitbang_device_init()
        if button not in self._button_map:
            raise ValueError("Invalid button {} specified".format(button))
        elif wait < 0.0:
            raise ValueError("Invalid wait {} specified".format(wait))
        elif not self._button_down[button]:
            self._gpio(button, -1, active_high=self._polarity)
            time.sleep(wait)

    def click(self, button, duration=.5):
        """Presses the button specified and holds it for the specified duration.

        Args:
            button (str): button to click.
            duration (float): seconds to wait before releasing the button.

        Raises:
            ValueError: invalid button or duration given.
        """
        if not self._bitbang_device_init:
            self.bitbang_device_init()
        if button not in self._button_map:
            raise ValueError("Invalid button {} specified".format(button))
        elif duration <= 0.0:
            raise ValueError("Invalid duration {} specified".format(duration))
        else:
            self._gpio(button, duration, active_high=self._polarity)

    def release(self, button):
        """Releases the button specified if it was previously pressed.

        Args:
            button (str): button to click.

        Raises:
            ValueError: invalid button given.
        """
        if not self._bitbang_device_init:
            self.bitbang_device_init()
        if button not in self._button_map:
            raise ValueError("Invalid button {} specified".format(button))
        elif self._button_down[button]:
            self._gpio(button, -1, active_high=self._polarity)

    def read_pin(self, interface, pin):
        """Returns True if the bit is high for the given pin, False otherwise.

        Args:
            interface (int): the port on the FTDI chip containing the GPIO
            pin (int): the bit representing the GPIO to read

        Return:
           bool: True if pin is high, False if low.
        """
        if not self._bitbang_device_init:
            self.bitbang_device_init()
        return (self._bb_map[interface].port & (1 << pin)) != 0

    def valid_buttons(self):
        """Returns a list of valid button names.

        Returns:
            list: A list of valid button name strings
        """
        return list(self._button_map.keys())

    def _gpio(self, button, duration, active_high):
        """Drive the GPIO of the button specified for the duration provided.

        Args:
            button (str): button to be toggled
            duration (float): seconds to wait between the two toggles. If negative, only one toggle
                      will be performed (either a press or release).
            active_high (bool): drive the pin high if True otherwise drive active_low.
        """
        if not self._bitbang_device_init:
            self.bitbang_device_init()
        interface = self._button_map[button]["interface"]
        bb = self._bb_map[interface]
        self.__toggle(button, bb, active_high)
        if duration >= 0.0:
            time.sleep(duration)
            self.__toggle(button, bb, active_high)

    def __toggle(self, button, bb, active_high):
        """Toggles the button specified based on its previous state.

        Args:
            button (str): button to be toggled
            bb (BitBangDevice): performs the toggle.
            active_high (bool): drive the pin high if True, otherwise drive active_low.
        """
        if not self._bitbang_device_init:
            self.bitbang_device_init()
        pin = self._button_map[button]["pin"]
        if self._button_down[button]:
            # Release button
            if active_high:
                self.__clear_pin(bb, pin)
            else:
                self.__set_pin(bb, pin)
            self._button_down[button] = False
        else:
            # Set button
            if active_high:
                self.__set_pin(bb, pin)
            else:
                self.__clear_pin(bb, pin)
            self._button_down[button] = True

    @staticmethod
    def __clear_pin(bb, pin):
        """Clear the bit of the given pin.

        Args:
            bb (BitBangDevice): performs the work
            pin (int): the bit representing the GPIO to clear
        """
        bb.port &= 255 ^ (1 << pin)

    @staticmethod
    def __set_pin(bb, pin):
        """Sets the bit for the given pin.

        Args:
            bb (BitBangDevice): performs the work
            pin (int): the bit representing the GPIO to clear
        """
        bb.port |= (1 << pin)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pseudoterminal transport class.

Some processes like the simulator shell script can fork/exec other
processes and need job control. A common way of overcoming this problem is
for the shell script to become a process leader. This is usually achieved by
using pseudoterminals.

More details: https://goo.gl/ckgF91

Pseudoterminals have a slightly different interface to reading and writing.

Example:
    os.read(self._process.stdout, size) in PTY
    self._process.stdout.read(size) in a regular process

This class will act as an adapter

"""
from __future__ import absolute_import
import os
import pty
import six
import subprocess
import time
from gazoo_device.switchboard import transport_properties
from gazoo_device.switchboard import process_transport
from gazoo_device import gdm_logger

logger = gdm_logger.get_gdm_logger()


class PtyTransport(process_transport.ProcessTransport):
    """Perform process transport over pseudoterminal."""

    def __init__(self,
                 comms_address,
                 args="",
                 auto_reopen=False,
                 open_on_start=True):
        """Initialize the pty transport object with the given process properties.

        Args:
            comms_address (str): the process command and args to communicate
            args (str): additional args to pass to the main process to open communication.
            auto_reopen (bool): flag indicating transport should be reopened if
                                unexpectedly closed.
            open_on_start (bool): flag indicating transport should be open on
                                  TransportProcess start.
        """
        self.comms_address = comms_address
        working_directory = os.path.dirname(comms_address)
        if working_directory:
            comms_address = comms_address.replace(working_directory, "./")
        else:
            working_directory = None
        super(PtyTransport, self).__init__(command=comms_address,
                                           args=args,
                                           auto_reopen=auto_reopen,
                                           open_on_start=open_on_start,
                                           working_directory=working_directory)
        self.primary = None
        self.secondary = None

    def close(self):
        """Closes the process if open.
        """
        if self.primary:
            os.close(self.primary)
            self.primary = None
        if self.secondary:
            os.close(self.secondary)
            self.secondary = None
        if hasattr(self, "_process") and self._process:
            if self._process.poll() is None:
                self._process.terminate()
                while self._process.poll() is None:
                    time.sleep(0.001)
            self._process = None

    def _open(self):
        """Opens or reopens the process using the current property values.

        Raises:
            Exception: when invalid executable or options are provided.
        """
        if self._is_ready_to_open():
            try:
                # self.primary corresponds to the "master" file descriptor from pty.openpty().
                # self.secondary corresponds to the "slave" file descriptor from pty.openpty().
                self.primary, self.secondary = pty.openpty()
                self._process = subprocess.Popen(  # pylint: disable=subprocess-popen-preexec-fn
                    self._args,
                    stdin=self.secondary,
                    stdout=self.secondary,
                    stderr=self.secondary,
                    cwd=self._cwd,
                    close_fds=self._properties[transport_properties.CLOSE_FDS],
                    preexec_fn=os.setsid  # This creates a new "process group"
                )
                # Set the stdin and stdout to self.primary
                self._process.stdout = self.primary
                self._process.stdin = self.primary
                os.close(self.secondary)
                self.secondary = None
            except Exception:
                self.close()
                raise

    def _read_data(self, size):
        try:
            return os.read(self._process.stdout, size)
        except OSError as err:
            logger.debug("Error in pty_process: {!r}".format(err))
            return b""

    def _write_data(self, data):
        if isinstance(data, six.integer_types):
            data = six.int2byte(data)
        os.write(self._process.stdin, data)
        return len(data)
